#: L23 Module Commands

from dataclasses import dataclass
import typing
import functools

from ..protocol.command_builders import (
    build_get_request,
    build_set_request
)
from .. import interfaces
from ..transporter.token import Token
from ..protocol.fields import data_types as xt
from ..protocol.fields.field import XmpField
from ..registry import register_command
from .enums import *  # noqa: F403
from xoa_driver.internals import warn


@register_command
@dataclass
class M_RESERVATION:
    """
    Set this command to reserve, release, or relinquish a module itself (as
    opposed to its ports). The module must be reserved before its hardware image can
    be upgraded. The owner of the session must already have been specified.
    Reservation will fail if the chassis or any ports are reserved for other users.

    .. note::

        The reservation parameters are slightly asymmetric with respect to set/get. When querying for the current reservation state, the chassis will use these values.

    """

    code: typing.ClassVar[int] = 72
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        operation: XmpField[xt.XmpByte] = XmpField(
            xt.XmpByte, choices=ReservedAction
        )  # coded byte, containing the operation to perform. The reservation parameters are asymmetric with respect to set/get. When set, it contains the operation to perform. When get, it contains the status.

    @dataclass(frozen=True)
    class GetDataAttr:
        operation: XmpField[xt.XmpByte] = XmpField(
            xt.XmpByte, choices=ReservedStatus
        )  # coded byte, containing the operation to perform. The reservation parameters are asymmetric with respect to set/get. When set, it contains the operation to perform. When get, it contains the status.

    def get(self) -> "Token[GetDataAttr]":
        """Get the reservation status of the test module.

        :return: the reservation status of the test module
        :rtype: M_RESERVATION.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, operation: ReservedAction) -> "Token":
        """Set the reservation status of the test module.

        :param operation: reservation operation to perform
        :type operation: ReservedAction
        """
        return Token(self._connection, build_set_request(self, module=self._module, operation=operation))

    set_release = functools.partialmethod(set, ReservedAction.RELEASE)
    """Release the test module.
    """
    set_reserve = functools.partialmethod(set, ReservedAction.RESERVE)
    """Reserve the test module.
    """
    set_relinquish = functools.partialmethod(set, ReservedAction.RELINQUISH)
    """Release the ownership of the test module from another user.
    """


@register_command
@dataclass
class M_RESERVEDBY:
    """
    Identify the user who has a module reserved. Returns an empty string if the
    module is not currently reserved by anyone.
    """

    code: typing.ClassVar[int] = 73
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        username: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, containing the name of the current owner of the module.

    def get(self) -> "Token[GetDataAttr]":
        """Get the username who has reserved the test module.

        :return: the username who has reserved the test module
        :rtype: M_RESERVEDBY.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_MODEL:
    """
    Gets the legacy model P/N name of a Xena test module.
    """

    code: typing.ClassVar[int] = 75
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        model: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the legacy model P/N name of a Xena test module.

    def get(self) -> "Token[GetDataAttr]":
        """Gets the legacy model P/N name of a Xena test module.

        :return: the legacy model P/N name of a Xena test module
        :rtype: M_MODEL.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_SERIALNO:
    """
    Gets the unique serial number of a module.
    """

    code: typing.ClassVar[int] = 76
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        serial_number: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the serial number of this module.

    def get(self) -> "Token[GetDataAttr]":
        """Gets the unique serial number of the test module.

        :return: the serial number of this test module
        :rtype: M_SERIALNO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_VERSIONNO:
    """
    Gets the version number of the hardware image installed on a module.
    """

    code: typing.ClassVar[int] = 77
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        version: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the hardware image version number.

    def get(self) -> "Token[GetDataAttr]":
        """Gets the version number of the hardware image installed on the test module.

        :return: the hardware image version number of the test module
        :rtype: M_VERSIONNO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_STATUS:
    """
    Get status readings for the test module itself.
    """

    code: typing.ClassVar[int] = 79
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        temperature: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, temperature of the main hardware chip, in degrees Celsius.

    def get(self) -> "Token[GetDataAttr]":
        """Get the status readings of the test module

        :return: temperature of the main hardware chip, in degrees Celsius
        :rtype: M_STATUS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_PORTCOUNT:
    """
    Gets the maximum number of ports on a module.

    .. note::

        For a CFP-type module this number refers to the maximum number of ports possible on the module regardless of the media configuration. So if a CFP-type module can be set in for instance either 1x100G mode or 8x10G mode then this command will always return 8. If you want the current number of ports for a CFP-type module you need to read the M_CFPCONFIGEXT` command which returns the number of current ports.

    """

    code: typing.ClassVar[int] = 80
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        port_count: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the maximum number of ports.

    def get(self) -> "Token[GetDataAttr]":
        """Gets the maximum number of ports on a module.

        :return: the maximum number of ports on the test module
        :rtype: M_PORTCOUNT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_UPGRADE:
    """
    Transfers a hardware image file from the chassis to a module. This image will
    take effect when the chassis is powered-on the next time. The transfer takes
    approximately 3 minutes, but no further action is required by the client.
    """

    code: typing.ClassVar[int] = 81
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        magic: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, must be the special value -1480937026.
        image_name: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the fully qualified name of a file previously uploaded to the chassis.

    def set(self, image_name: str) -> "Token":
        """Transfers a hardware image file from the chassis to a module. This image will
        take effect when the chassis is powered-on the next time. The transfer takes
        approximately 3 minutes, but no further action is required by the client.

        :param image_name: the fully qualified name of a file previously uploaded to the chassis
        :type image_name: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, magic=-1480937026, image_name=image_name))


@register_command
@dataclass
class M_UPGRADEPROGRESS:
    """
    Provides a value indicating the current stage of an ongoing hardware image
    upgrade operation. This is for information only; the upgrade operation runs to
    completion by itself. The progress values are pushed to the client without it
    having to request them.
    """

    code: typing.ClassVar[int] = 82
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        progress: XmpField[xt.XmpInt] = XmpField(
            xt.XmpInt
        )  # integer, the current stage within the three phases. 0: Failure. 1-100: Erase completion percentage. 101-200: Write completion percentage. 201-300: Verify completion percentage.

    def get(self) -> "Token[GetDataAttr]":
        """Get the current stage of an ongoing hardware image
        upgrade operation. This is for information only; the upgrade operation runs to
        completion by itself. The progress values are pushed to the client without it
        having to request them.

        :param progress: the current stage within the three phases. 0: Failure. 1-100: Erase completion percentage. 101-200: Write completion percentage. 201-300: Verify completion percentage.
        :type progress: M_UPGRADEPROGRESS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_TIMESYNC:
    """
    Control how the test module timestamp clock is running, either freely in the
    chassis or locked to an external system time. Running with free chassis time
    allows nano-second precision measurements of latencies, but only when the
    transmitting and receiving ports are in the same chassis. Running with locked
    external time enables inter-chassis latency measurements, but can introduce
    small time discontinuities as the test module time is adjusted.
    """

    code: typing.ClassVar[int] = 83
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        source: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=TimingSource)  # coded byte, selecting the time sync mode.

    @dataclass(frozen=True)
    class GetDataAttr:
        source: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=TimingSource)  # coded byte, selecting the time sync mode.

    def get(self) -> "Token[GetDataAttr]":
        """Get the time sync mode of the test module timestamp clock.

        :return: the time sync mode of the test module timestamp clock
        :rtype: M_TIMESYNC.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, source: TimingSource) -> "Token":
        """Set the time sync mode of the test module timestamp clock.

        :param mode: the time sync mode of the test module timestamp clock
        :type mode: TimingSource
        """
        return Token(self._connection, build_set_request(self, module=self._module, source=source))

    set_chassis = functools.partialmethod(set, TimingSource.CHASSIS)
    """Set the time sync mode of the test module to Chassis Mode.
    """
    set_external = functools.partialmethod(set, TimingSource.EXTERNAL)
    """Set the time sync mode of the test module to External Mode.
    """
    set_module = functools.partialmethod(set, TimingSource.MODULE)
    """Set the time sync mode of the test module to Module Mode.
    """


@register_command
@dataclass
class M_CFPTYPE:
    """
    Get information about the transceiver currently inserted into the cages.
    """

    code: typing.ClassVar[int] = 84
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        state: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=MediaCFPState)  # coded byte, specifying the CFP state.
        type: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=MediaCFPType)  # coded byte, specifying the CFP type.

    def get(self) -> "Token[GetDataAttr]":
        """Get CFP type information about the transceiver currently inserted into the cage.

        :return:
            - the CFP state
            - the CFP type
        :rtype: M_CFPTYPE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_CFPCONFIG:
    """
    .. deprecated:: 1.3

    The current number of ports and their speed of a CFP test module. If the CFP
    type is NOTFLEXIBLE then it reflects the transceiver currently in the CFP cage.
    If the CFP type is FLEXIBLE (or NOTPRESENT) then the configuration can be changed
    explicitly. The following combinations are possible: 4x10G, 8x10G, 1x40G, 2x40G,
    and 1x100G. (replaced by :class:`M_CFPCONFIGEXT`)
    """

    code: typing.ClassVar[int] = 85
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        port_count: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)  # byte, number of ports.
        port_speed: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)  # byte, port speed, in Gbps.

    @dataclass(frozen=True)
    class GetDataAttr:
        port_count: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)  # byte, number of ports.
        port_speed: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)  # byte, port speed, in Gbps.

    def get(self) -> "Token[GetDataAttr]":
        """Get the current number of ports and their speed of a CFP test module.

        :return:
            - number of ports
            - port speed, in Gbps
        :rtype: M_CFPCONFIG.GetDataAttr
        """

        warn.depricated("module.cfp.config.get() (M_CFPCONFIG) is deprecated. Please use module.cfp.config_extended.get() (M_CFPCONFIGEXT) instead.")

        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, port_count: int, port_speed: int) -> "Token":
        """Set the current number of ports and their speed of a CFP test module.

        :param port_count: number of ports
        :type port_count: int
        :param port_speed: port speed, in Gbps
        :type port_speed: int
        """

        warn.depricated("module.cfp.config.set() (M_CFPCONFIG) is deprecated. Please use module.cfp.config_extended.set() (M_CFPCONFIGEXT) instead.")

        return Token(self._connection, build_set_request(self, module=self._module, port_count=port_count, port_speed=port_speed))


@register_command
@dataclass
class M_COMMENT:
    """
    Gets the user-defined description string of a module.
    """

    code: typing.ClassVar[int] = 86
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        comment: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the user-specified comment/description for the module.

    @dataclass(frozen=True)
    class GetDataAttr:
        comment: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the user-specified comment/description for the module.

    def get(self) -> "Token[GetDataAttr]":
        """Get the user-defined description string of a module.

        :return: the user-specified comment/description for the module
        :rtype: M_COMMENT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, comment: str) -> "Token":
        """Set the user-defined description string of a module.

        :param comment: the user-specified comment/description for the module
        :type comment: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, comment=comment))


@register_command
@dataclass
class M_TIMEADJUSTMENT:
    """
    Control time adjustment for module wall clock.
    """

    code: typing.ClassVar[int] = 88
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        adjust: XmpField[xt.XmpInt] = XmpField(
            xt.XmpInt
        )  # integer, adjustment in nanoseconds. This value should be a multiple of 8 as it will be converted to a number of 125 MHz clocks.

    @dataclass(frozen=True)
    class GetDataAttr:
        adjust: XmpField[xt.XmpInt] = XmpField(
            xt.XmpInt
        )  # integer, adjustment in nanoseconds. This value should be a multiple of 8 as it will be converted to a number of 125 MHz clocks.

    def get(self) -> "Token[GetDataAttr]":
        """Get the time adjustment value for the module clock.

        :return: the time adjustment value for the module clock
        :rtype: M_TIMEADJUSTMENT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, adjust: int) -> "Token":
        """Set the time adjustment value for the module clock. This value should be a multiple of 8 as it will be converted to a number of 125 MHz clocks.

        :param adjust: the time adjustment value for the module clock
        :type adjust: int
        """
        return Token(self._connection, build_set_request(self, module=self._module, adjust=adjust))


@register_command
@dataclass
class M_CAPABILITIES:
    """
    Gets the module capabilities.
    """

    code: typing.ClassVar[int] = 89
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        can_advanced_timing: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=YesNo)  # coded integer, is advanced timing functions supported?
        can_local_time_adjust: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=YesNo)  # coded integer, is local time adjustment supported?
        can_media_config: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=YesNo)  # coded integer, is module media configuration supported?
        require_multi_image: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=YesNo)  # coded integer, does this module switch images during runtime?
        is_chimera: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=YesNo)  # coded integer, is this a Chimera module?
        max_clock_ppm: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, maximum supported absolute +- clock ppm setting.
        can_tsn: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=YesNo)  # coded integer, does this module support Time Sensitive Networking (TSN) ?
        can_ppm_sweep: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=YesNo)  # coded integer, does this module support Local Clock Adjustment/Sweep (aka. PPM Sweep) ?

    def get(self) -> "Token[GetDataAttr]":
        """Get the test module capabilities.

        :return:
            - is advanced timing functions supported?
            - is local time adjustment supported?
            - is module media configuration supported?
            - does this module switch images during runtime?
            - is this a Chimera module?
            - maximum supported absolute +- clock ppm setting.
            - does this module support Time Sensitive Networking (TSN) ?
            - does this module support Local Clock Adjustment/Sweep (aka. PPM Sweep) ?

        :rtype: M_CAPABILITIES.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_MEDIASUPPORT:
    """
    This command shows the available speeds on a module. The structure of the returned value is
    ``[ <cage_type> <available_speed_count> [<ports_per_speed> <speed>] ]``.
    ``[<ports_per_speed> <speed>]`` is repeated until all speeds supported by the ``<cage_type>`` has been listed.
    ``[<cage_type> <available_speed_count>]`` is repeated for all cage types on the module including the related ``<ports_per_speed> <speed>`` information.
    """

    code: typing.ClassVar[int] = 90
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        media_info_list: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)  # coded integer, media information

    def get(self) -> "Token[GetDataAttr]":
        """Get the media supports by the port, including cage type, available speed count, ports per speed, and the corresponding speed.

        :return:
            a list of integers. The structure of the returned value is ``[ <cage_type> <available_speed_count>[<ports_per_speed> <speed>] ]``.
            ``[<ports_per_speed> <speed>]`` is repeated until all speeds supported by the ``<cage_type>`` has been listed.
            ``[<cage_type> <available_speed_count>]`` is repeated for all cage types on the module including the related ``<ports_per_speed> <speed>`` information.

        :rtype: M_MEDIASUPPORT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_FPGAREIMAGE:
    """
    Reload FPGA image.
    """

    code: typing.ClassVar[int] = 91
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        key_code: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, must be 42.

    def set(self) -> "Token":
        """Reload the FPGA image.

        :param key_code: must be 42.
        :type key_code: int.
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                key_code=42,
            ),
        )


@register_command
@dataclass
class M_MULTIUSER:
    """
    Enable or disable multiple sessions to control the same module.
    """

    code: typing.ClassVar[int] = 92
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, enable or disable multiple sessions to control the same module.

    @dataclass(frozen=True)
    class GetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, enable or disable multiple sessions to control the same module.

    def get(self) -> "Token[GetDataAttr]":
        """Get the status of multiple sessions controlling the same module.

        :return: the status of multiple sessions controlling the same module
        :rtype: M_MULTIUSER.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, on_off: OnOff) -> "Token":
        """Enable or disable multiple sessions to control the same module.

        :param on_off: Enable or disable multiple sessions to control the same module
        :type on_off: OnOff
        """
        return Token(self._connection, build_set_request(self, module=self._module, on_off=on_off))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Disable multiple sessions to control the same module.
    """
    set_on = functools.partialmethod(set, OnOff.ON)
    """Enable multiple sessions to control the same module.
    """


@register_command
@dataclass
class M_CFPCONFIGEXT:
    """
    This property defines the current number of ports and the speed of each of them
    on a CFP test module. If the CFP type is ``NOTFLEXIBLE`` then it reflects the
    transceiver currently in the CFP cage. If the CFP type is ``FLEXIBLE`` (or
    ``NOTPRESENT``) then the configuration can be changed explicitly. The following
    combinations are possible: 2x10G, 4x10G, 8x10G, 2x25G, 4x25G, 8x25G, 1x40G,
    2x40G, 2x50G, 4x50G, 8x50G, 1x100G, 2x100G, 4x100G, 2x200G, and 1x400G.
    (replaces :class:`M_CFPCONFIG`)

    .. note::

        ``<portspeed_list>`` is a list of integers, where the first element is the number of ports followed by a number of port speeds in Mbps.
        The number of port speeds equals the value of the number of ports.
        For example if the configuration is 4x25G, ``<portspeed_list>`` will be ``[4, 25000, 25000, 25000, 25000]``.
    """

    code: typing.ClassVar[int] = 93
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        portspeed_list: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)

    @dataclass(frozen=True)
    class GetDataAttr:
        portspeed_list: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)

    def get(self) -> "Token[GetDataAttr]":
        """Get a list of port count and corresponding speeds supported by the current module config.

        :return: a list of port count and corresponding speeds supported by the current module config
        :rtype: M_CFPCONFIGEXT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, portspeed_list: typing.List[int]) -> "Token":
        return Token(self._connection, build_set_request(self, module=self._module, portspeed_list=portspeed_list))


@register_command
@dataclass
class M_CLOCKPPB:
    """
    Makes small adjustments to the local clock of the test module, which drives the
    TX rate of the test ports.
    """

    code: typing.ClassVar[int] = 94
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        ppb: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, adjustment from nominal value, in parts-per-billion, positive or negative.

    @dataclass(frozen=True)
    class GetDataAttr:
        ppb: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, adjustment from nominal value, in parts-per-billion, positive or negative.

    def get(self) -> "Token[GetDataAttr]":
        """Get the module clock adjustment in ppb.

        :return: the module clock adjustment in ppb
        :rtype: M_CLOCKPPB.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, ppb: int) -> "Token":
        """Set the module clock adjustment in ppb.

        :param ppb: adjustment from nominal value, in parts-per-billion, positive or negative
        :type ppb: int
        """
        return Token(self._connection, build_set_request(self, module=self._module, ppb=ppb))


@register_command
@dataclass
class M_SMAINPUT:
    """
    For test modules with SMA (SubMiniature version A) connectors, selects the function of the SMA input.
    """

    code: typing.ClassVar[int] = 95
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        sma_in: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=SMAInputFunction)  # coded byte, specifying the function of the SMA input.

    @dataclass(frozen=True)
    class GetDataAttr:
        sma_in: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=SMAInputFunction)  # coded byte, specifying the function of the SMA input.

    def get(self) -> "Token[GetDataAttr]":
        """Get the function of the SMA (SubMiniature version A) input of the module

        :return: the function of the SMA (SubMiniature version A) input of the module
        :rtype: M_SMAINPUT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, sma_in: SMAInputFunction) -> "Token":
        """Set the function of the SMA (SubMiniature version A) input of the module

        :param sma_in: the function of the SMA (SubMiniature version A) input of the module
        :type sma_in: SMAInputFunction
        """
        return Token(self._connection, build_set_request(self, module=self._module, sma_in=sma_in))

    set_notused = functools.partialmethod(set, SMAInputFunction.NOT_USED)
    """Set SMA input to Not Used
    """
    set_tx2mhz = functools.partialmethod(set, SMAInputFunction.TX2MHZ)
    """Set SMA input to TX Clock Ref. 2.048 MHz
    """
    set_tx10mhz = functools.partialmethod(set, SMAInputFunction.TX10MHZ)
    """Set SMA input to TX Clock Ref. 10.0 MHz
    """


@register_command
@dataclass
class M_SMAOUTPUT:
    """
    For test modules with SMA (SubMiniature version A) connectors, selects the function of the SMA output.
    """

    code: typing.ClassVar[int] = 96
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        sma_out: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=SMAOutputFunction)  # coded byte, specifying the function of the SMA output.

    @dataclass(frozen=True)
    class GetDataAttr:
        sma_out: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=SMAOutputFunction)  # coded byte, specifying the function of the SMA output.

    def get(self) -> "Token[GetDataAttr]":
        """Get the function of the SMA (SubMiniature version A) output of the module

        :return: the function of the SMA (SubMiniature version A) output of the module
        :rtype: M_SMAOUTPUT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, sma_out: SMAOutputFunction) -> "Token":
        """Set the function of the SMA (SubMiniature version A) output of the module

        :param sma_out: the function of the SMA (SubMiniature version A) output of the module
        :type sma_out: SMAOutputFunction
        """
        return Token(self._connection, build_set_request(self, module=self._module, sma_out=sma_out))

    set_disabled = functools.partialmethod(set, SMAOutputFunction.DISABLED)
    """Set SMA output function to Disabled.
    """
    set_passthrough = functools.partialmethod(set, SMAOutputFunction.PASSTHROUGH)
    """Set SMA output function to Pass-Through.
    """
    set_p0sof = functools.partialmethod(set, SMAOutputFunction.P0SOF)
    """Set SMA output function to Port 0 Start-of-Frame Pulse.
    """
    set_p1sof = functools.partialmethod(set, SMAOutputFunction.P1SOF)
    """Set SMA output function to Port 1 Start-of-Frame Pulse.
    """
    set_ref2mhz = functools.partialmethod(set, SMAOutputFunction.REF2MHZ)
    """Set SMA output function to TX Clock (nom. 2.048 MHz).
    """
    set_ref10mhz = functools.partialmethod(set, SMAOutputFunction.REF10MHZ)
    """Set SMA output function to TX Clock (nom. 10.0 MHz).
    """
    set_ref125mhz = functools.partialmethod(set, SMAOutputFunction.REF125MHZ)
    """Set SMA output function to TX Clock (nom. 125 MHz).
    """
    set_ref156mhz = functools.partialmethod(set, SMAOutputFunction.REF156MHZ)
    """Set SMA output function to TX Clock (nom. 156.25 MHz).
    """
    set_p0rxclk = functools.partialmethod(set, SMAOutputFunction.P0RXCLK)
    """Set SMA output function to Port 0 RX Clock (nom. 10.0 MHz).
    """
    set_p1rxclk = functools.partialmethod(set, SMAOutputFunction.P1RXCLK)
    """Set SMA output function to Port 1 RX Clock (nom. 10.0 MHz).
    """
    set_p0rxclk2mhz = functools.partialmethod(set, SMAOutputFunction.P0RXCLK2MHZ)
    """Set SMA output function to Port 0 RX Clock (nom. 2.048 MHz).
    """
    set_p1rxclk2mhz = functools.partialmethod(set, SMAOutputFunction.P1RXCLK2MHZ)
    """Set SMA output function to Port 1 RX Clock (nom. 2.048 MHz).
    """
    set_ts_pps = functools.partialmethod(set, SMAOutputFunction.TS_PPS)
    """Set SMA output function to Timing Source (Pulse-Per-Second).
    """


@register_command
@dataclass
class M_SMASTATUS:
    """
    For test modules with SMA connectors, this returns the status of the SMA input.
    """

    code: typing.ClassVar[int] = 97
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        status: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=SMAStatus)  # coded byte, specifying the status of the SMA input.

    def get(self) -> "Token[GetDataAttr]":
        """Get the status of the SMA input

        :return: the status of the SMA input
        :rtype: M_SMASTATUS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_NAME:
    """
    Gets the name of a module.
    """

    code: typing.ClassVar[int] = 99
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        name: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the name for the module.

    def get(self) -> "Token[GetDataAttr]":
        """Get the name of the module.

        :return: the name of the module
        :rtype: M_NAME.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_REVISION:
    """
    Gets the model P/N name of a Xena test module.
    """

    code: typing.ClassVar[int] = 100
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        revision: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the model P/N name of a Xena test module.

    def get(self) -> "Token[GetDataAttr]":
        """Get the model P/N name of a Xena test module.

        :return: the model P/N name of a Xena test module.
        :rtype: M_REVISION.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_MEDIA:
    """
    For the test modules that support media configuration (check M_CAPABILITIES), this command sets the desired media
    type (front port).
    """

    code: typing.ClassVar[int] = 342
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        media_config: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=MediaConfigurationType)  # coded byte, specifying the active front port: CFP4, QSFP28, CXP, SFP28.

    @dataclass(frozen=True)
    class GetDataAttr:
        media_config: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=MediaConfigurationType)  # coded byte, specifying the active front port: CFP4, QSFP28, CXP, SFP28.

    def get(self) -> "Token[GetDataAttr]":
        """Get the media type of the test module.

        :return: the media type of the test module
        :rtype: M_MEDIA.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, media_config: MediaConfigurationType) -> "Token":
        """Set the media type of the test module.

        :param media_config: the media type of the test module
        :type media_config: MediaType
        """
        return Token(self._connection, build_set_request(self, module=self._module, media_config=media_config))


@register_command
@dataclass
class M_CLOCKSYNCSTATUS:
    """
    Get module's clock sync status.
    """

    code: typing.ClassVar[int] = 370
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        m_clock_diff: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, module clock diff
        m_correction: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, module correction
        m_tune_is_increase: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, whether module tune is increased
        m_tune_value: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, module tune value
        m_is_steady_state: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, whether module is in steady state

    def get(self) -> "Token[GetDataAttr]":
        """Get the test module's clock sync status.

        :return: the test module's clock sync status
        :rtype: M_CLOCKSYNCSTATUS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_LICENSE_DEMO_INFO:
    """
    Returns info about the demo status of the module. Only applicable to L47 test module.
    """

    code: typing.ClassVar[int] = 400
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        demo: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=HasDemo)  # coded byte, specifies if this is a demo module or not.
        valid: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=IsValid)  # coded byte, if this is a demo module, specifies if the demo license is valid.
        permanent: XmpField[xt.XmpByte] = XmpField(
            xt.XmpByte, choices=IsPermanent
        )  # coded byte, if this is a demo module and the demo license is valid, specifies if the demo license is permanent.
        expiration: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long integer, if this is a demo module and the demo license is valid and not permanent, specifies the expiration date of the demo license - in seconds since Jan 1, 1970.

    def get(self) -> "Token[GetDataAttr]":
        """Get info of the demo status of the test module. Only applicable to L47 test module.

        :return: info of the demo status of the test module.
        :rtype: M_LICENSE_DEMO_INFO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_LICENSE_MAINTENANCE_INFO:
    """
    Returns info about the maintenance license status for the module. Only applicable to L47 test module.
    """

    code: typing.ClassVar[int] = 401
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        valid: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=IsValid)  # coded byte, specifies if the maintenance license is valid.
        permanent: XmpField[xt.XmpByte] = XmpField(
            xt.XmpByte, choices=IsPermanent
        )  # coded byte, if the maintenance license is valid, specifies if the maintenance license is permanent.
        expiration: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long integer, if the maintenance license is valid and not permanent, specifies the expiration date of the maintenance license - in seconds since Jan 1, 1970.

    def get(self) -> "Token[GetDataAttr]":
        """Get the info about the maintenance license status for the module. Only applicable to L47 test module.

        :return: the info about the maintenance license status for the test module
        :rtype: M_LICENSE_MAINTENANCE_INFO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_LICENSE_CWB_DETECTED:
    """
    Returns if clock-windback is detected. If clock-windback has been detected the
    chassis is locked and no reservations of ports can be performed. To recover from
    clock-windback, set the system time correct (via the M4_SYSTEM_TIME command) and
    perform a license update (via the M_LICENSE_UPDATE command). Only applicable to L47 test module.
    """

    code: typing.ClassVar[int] = 402
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        detected: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=YesNo)  # coded byte, specifies if clock-windback is detected.

    def get(self) -> "Token[GetDataAttr]":
        """Get whether clock-windback is detected.

        :return: whether clock-windback is detected
        :rtype: M_LICENSE_CWB_DETECTED.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_LICENSE_UPDATE:
    """
    This command instructs the chassis to update its local license information. The chassis can be configured in on-line and off-line mode
    (by the M_LICENSE_ONLINE command). In on-line mode, the chassis sends a
    capability request and receives a capability response. In
    offline mode a capability response (bin file) must be downloaded and uploaded to the chassis. The capability response (bin file) is
    parsed and the license info is stored locally in trusted storage. A capability
    response (bin file) has a lifetime of one day (24 hours). The result of the
    license update operation can be retrieved by M_LICENSE_UPDATE_STATUS.
    """

    code: typing.ClassVar[int] = 403
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        pass

    def set(self) -> "Token":
        """Start license update
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
            ),
        )


@register_command
@dataclass
class M_LICENSE_UPDATE_STATUS:
    """
    Returns the status of the latest license update operations.
    """

    code: typing.ClassVar[int] = 404
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        update_state: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=UpdateState)  # coded byte, specifies the state of the license update procedure
        last_update: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, time for the last update request - in seconds since Jan 1, 1979
        last_success: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, time for the last successful update - in seconds since Jan 1, 1979
        last_fail: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, time for the last failed update - in seconds since Jan 1, 1979
        Info: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, info about the last license update operation - reason for failed update.

    def get(self) -> "Token[GetDataAttr]":
        """Get the status of the latest license update operation.

        :return: the status of the latest license update operation
        :rtype: M_LICENSE_UPDATE_STATUS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_LICENSE_LIST_BSON:
    """
    Returns a list of locally stored licenses - formatted as a BSON document.
    """

    code: typing.ClassVar[int] = 405
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        bson: XmpField[xt.XmpHexList] = XmpField(xt.XmpHexList)  # list of hex bytes, bson document containing the list of locally stored licenses

    def get(self) -> "Token[GetDataAttr]":
        """Get the a list of locally stored licenses - formatted as a BSON document.

        :return: a list of locally stored licenses - formatted as a BSON document.
        :rtype: M_LICENSE_LIST_BSON.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_LICENSE_ONLINE:
    """
    Configures the chassis in online or offline mode. The online mode configuration
    defines two different license update procedures as described for the
    M_LICENSE_UPDATE command. In online mode the license update procedure requires
    access to the Internet. In offline mode the license update procedure can be
    performed without access to the Internet.
    """

    code: typing.ClassVar[int] = 406
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        mode: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=IsOnline)  # coded byte, chassis online/offline mode.

    @dataclass(frozen=True)
    class GetDataAttr:
        mode: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=IsOnline)  # coded byte, chassis online/offline mode.

    def get(self) -> "Token[GetDataAttr]":
        """Get the current online/offline mode of the L47 tester.

        :return: the current online/offline mode of the L47 tester
        :rtype: M_LICENSE_ONLINE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, mode: IsOnline) -> "Token":
        """Set the current online/offline mode of the L47 tester.

        :param mode: the current online/offline mode of the L47 tester
        :type mode: IsOnline
        """
        return Token(self._connection, build_set_request(self, module=self._module, mode=mode))

    set_offline = functools.partialmethod(set, IsOnline.OFFLINE)
    """Set the L47 tester to offline mode.
    """
    set_online = functools.partialmethod(set, IsOnline.ONLINE)
    """Set the L47 tester to online mode.
    """


@register_command
@dataclass
class M_TXCLOCKSOURCE_NEW:
    """
    For test modules with advanced timing features, select what clock drives the port TX
    rates.
    """

    code: typing.ClassVar[int] = 410
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        tx_clock: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=TXClockSource)  # coded byte, specifying what drives the port TX rates.

    @dataclass(frozen=True)
    class GetDataAttr:
        tx_clock: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=TXClockSource)  # coded byte, specifying what drives the port TX rates.

    def get(self) -> "Token[GetDataAttr]":
        """Get the test module's TX clock source settings.

        :return: the test module's TX clock source settings.
        :rtype: M_TXCLOCKSOURCE_NEW.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, tx_clock: TXClockSource) -> "Token":
        """Set the test module's TX clock source settings.

        :param tx_clock: the test module's TX clock source settings
        :type tx_clock: TXClockSource
        """
        return Token(self._connection, build_set_request(self, module=self._module, tx_clock=tx_clock))

    set_modulelocalclock = functools.partialmethod(set, TXClockSource.MODULELOCALCLOCK)
    """Set the test module's TX clock source to Module Local Clock
    """
    set_smainput = functools.partialmethod(set, TXClockSource.SMAINPUT)
    """Set the test module's TX clock source to SMA Input
    """
    set_p0rxclk = functools.partialmethod(set, TXClockSource.P0RXCLK)
    """Set the test module's TX clock source to Port 0 RX Clock
    """
    set_p1rxclk = functools.partialmethod(set, TXClockSource.P1RXCLK)
    """Set the test module's TX clock source to Port 1 RX Clock
    """
    set_p2rxclk = functools.partialmethod(set, TXClockSource.P2RXCLK)
    """Set the test module's TX clock source to Port 2 RX Clock
    """
    set_p3rxclk = functools.partialmethod(set, TXClockSource.P3RXCLK)
    """Set the test module's TX clock source to Port 3 RX Clock
    """
    set_p4rxclk = functools.partialmethod(set, TXClockSource.P4RXCLK)
    """Set the test module's TX clock source to Port 4 RX Clock
    """
    set_p5rxclk = functools.partialmethod(set, TXClockSource.P5RXCLK)
    """Set the test module's TX clock source to Port 5 RX Clock
    """
    set_p6rxclk = functools.partialmethod(set, TXClockSource.P6RXCLK)
    """Set the test module's TX clock source to Port 6 RX Clock
    """
    set_p7rxclk = functools.partialmethod(set, TXClockSource.P7RXCLK)
    """Set the test module's TX clock source to Port 7 RX Clock
    """


@register_command
@dataclass
class M_TXCLOCKSTATUS_NEW:
    """
    For test modules with advanced timing features, check whether a valid clock is
    present.
    """

    code: typing.ClassVar[int] = 411
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        status: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=TXClockStatus)  # coded byte, specifying the status of the TX clock.

    def get(self) -> "Token[GetDataAttr]":
        """Get the status of whether a valid clock is present for the test module.

        :return: the status of whether a valid clock is present for the test module.
        :rtype: M_TXCLOCKSTATUS_NEW.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_TXCLOCKFILTER_NEW:
    """
    For test modules with advanced timing features, the loop bandwidth on the TX
    clock filter.
    """

    code: typing.ClassVar[int] = 412
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        filter_bandwidth: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=LoopBandwidth)  # coded byte, the loop bandwidth on the TX clock filter.

    @dataclass(frozen=True)
    class GetDataAttr:
        filter_bandwidth: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=LoopBandwidth)  # coded byte, the loop bandwidth on the TX clock filter.

    def get(self) -> "Token[GetDataAttr]":
        """Get the setting of the loop bandwidth on the TX clock filter.

        :return: the setting of the loop bandwidth on the TX clock filter.
        :rtype: M_TXCLOCKFILTER_NEW.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, filter_bandwidth: LoopBandwidth) -> "Token":
        """Set the setting of the loop bandwidth on the TX clock filter.

        :param filter_bandwidth: the setting of the loop bandwidth on the TX clock filter
        :type filter_bandwidth: LoopBandwidth
        """
        return Token(self._connection, build_set_request(self, module=self._module, filter_bandwidth=filter_bandwidth))

    set_bw103hz = functools.partialmethod(set, LoopBandwidth.BW103HZ)
    """Set the loop bandwidth on the TX clock filter to BW = 103 Hz.
    """
    set_bw207hz = functools.partialmethod(set, LoopBandwidth.BW207HZ)
    """Set the loop bandwidth on the TX clock filter to BW = 207 Hz.
    """
    set_bw416hz = functools.partialmethod(set, LoopBandwidth.BW416HZ)
    """Set the loop bandwidth on the TX clock filter to BW = 416 Hz.
    """
    set_bw1683hz = functools.partialmethod(set, LoopBandwidth.BW1683HZ)
    """Set the loop bandwidth on the TX clock filter to BW = 1683 Hz.
    """
    set_bw7019hz = functools.partialmethod(set, LoopBandwidth.BW7019HZ)
    """Set the loop bandwidth on the TX clock filter to BW = 7019 Hz.
    """


@register_command
@dataclass
class M_CLOCKPPBSWEEP:
    """
    .. versionadded:: 1.1

    Start and stop deviation sweep the local clock of the test module, which drives the TX rate of the test ports.

    Note: The sweep is independent of the :class:`M_CLOCKPPB` parameter, i.e. the sweep uses the deviation set by :class:`M_CLOCKPPB` as its zero point.
    """

    code: typing.ClassVar[int] = 413
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        mode: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=PPMSweepMode)  # coded byte, specifying the sweeping function: OFF or TRIANGLE
        # integer >=0, the numeric clock adjustment in ppb per step of the sweep. If set to 0, the sweep will use as small steps as possible, creating a "linear" sweep of the clock rate.
        ppb_step: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)
        # integer >0 the delay in µs between each step in the sweep. If ppb_step is 0: The total time in µs to sweep linearly from 0 to max_ppb.
        step_delay: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)
        # integer != 0, the numeric maximum clock adjustment. The sign of max_ppb determines if the sweep will start with positive or negative offsets. When the next step would exceed the limit set by max_ppb, the sweep changes direction. I.e. the deviation will sweep from 0 to max_ppb, to (-max_ppb), and back to 0.
        max_ppb: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)
        loops: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer >=0, the number of full sweeps performed. 0 means "indefinitely".

    @dataclass(frozen=True)
    class GetDataAttr:
        mode: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=PPMSweepMode)  # coded byte, specifying the sweeping function.
        # integer >=0, the numeric clock adjustment in ppb per step of the sweep. If set to 0, the sweep will use as small steps as possible, creating a "linear" sweep of the clock rate.
        ppb_step: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)
        # integer >0 the delay in µs between each step in the sweep. If ppb_step is 0: The total time in µs to sweep linearly from 0 to max_ppb.
        step_delay: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)
        # integer != 0, the numeric maximum clock adjustment. The sign of max_ppb determines if the sweep will start with positive or negative offsets. When the next step would exceed the limit set by max_ppb, the sweep changes direction. I.e. the deviation will sweep from 0 to max_ppb, to (-max_ppb), and back to 0.
        max_ppb: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)
        loops: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer >=0, the number of full sweeps performed. 0 means "indefinitely".

    def get(self) -> "Token[GetDataAttr]":
        """Get the PPM sweep parameters from the module.

        :return: the PPM sweep parameters from the module.
        :rtype: M_CLOCKPPBSWEEP.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, mode: PPMSweepMode, ppb_step: int, step_delay: int, max_ppb: int, loops: int) -> "Token":
        """Set the PPM sweep parameters of the module.

        :param mode: specifying the sweeping function: OFF or TRIANGLE.
        :type mode: PPMSweepMode
        :param ppb_step: >=0, the numeric clock adjustment in ppb per step of the sweep. If set to 0, the sweep will use as small steps as possible, creating a "linear" sweep of the clock rate.
        :type ppb_step: int
        :param step_delay: >0 the delay in µs between each step in the sweep. If ppb_step is 0: The total time in µs to sweep linearly from 0 to max_ppb.
        :type step_delay: int
        :param max_ppb: != 0, the numeric maximum clock adjustment. The sign of max_ppb determines if the sweep will start with positive or negative offsets. When the next step would exceed the limit set by max_ppb, the sweep changes direction. I.e. the deviation will sweep from 0 to max_ppb, to (-max_ppb), and back to 0.
        :type max_ppb: int
        :param loops: >=0, the number of full sweeps performed. 0 means "indefinitely".
        :type loops: int
        """
        return Token(self._connection, build_set_request(self, module=self._module, mode=mode, ppb_step=ppb_step, step_delay=step_delay, max_ppb=max_ppb, loops=loops))


@register_command
@dataclass
class M_CLOCKSWEEPSTATUS:
    """
    .. versionadded:: 1.1

    Return the current status of the :class:`M_CLOCKPPBSWEEP` function.

    """

    code: typing.ClassVar[int] = 414
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        state: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=PPMSweepStatus)  # coded byte, specifying if a sweep is active: OFF or SWEEPING

        curr_sweep: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer >=0, the current full sweep number, counting from 0.
        curr_step: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer >=0 the current step number inside the sweep, counting from 0.

        # integer, >0, the total number of steps comprising a full sweep. For "linear" sweeps (ppb_step=0, see M_CLOCKPPBSWEEP) this number is determined by the chassis. In other cases, the number is implicitly given by the M_CLOCKPPBSWEEP parameters.
        max_steps: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)

    def get(self) -> "Token[GetDataAttr]":
        """Get the current status of the :class:`M_CLOCKPPBSWEEP` function.

        :return: the current status of the :class:`M_CLOCKPPBSWEEP` function.
        :rtype: M_CLOCKSWEEPSTATUS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M_LATENCYMODE:
    """
    Configures the latency mode for Chimera module. In extended latency mode, the FPGA allows all latency parameters to be 10 times higher, at the cost of reduced latency precision.

    .. note::

        - When change the latency mode, all latency configurations are reset on all ports in chimera module.

    """

    code: typing.ClassVar[int] = 450
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        mode: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=ImpairmentLatencyMode)  # coded byte, specifying latency mode.

    @dataclass(frozen=True)
    class GetDataAttr:
        mode: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=ImpairmentLatencyMode)  # coded byte, specifying latency mode.

    def get(self) -> "Token[GetDataAttr]":
        """Get the latency mode of the Chimera module.

        :return: the latency mode of the Chimera module.
        :rtype: M_LATENCYMODE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, mode: ImpairmentLatencyMode) -> "Token":
        """Set the latency mode of the Chimera module.

        :param mode: the bypass mode of the impairment emulator.
        :type mode: ImpairmentLatencyMode
        """
        return Token(self._connection, build_set_request(self, module=self._module, mode=mode))

    set_normal = functools.partialmethod(set, ImpairmentLatencyMode.NORMAL)
    """Set the latency mode of the Chimera module to NORMAL
    """
    set_extended = functools.partialmethod(set, ImpairmentLatencyMode.EXTENDED)
    """Set the latency mode of the Chimera module to EXTENDED
    """


@register_command
@dataclass
class M_EMULBYPASS:
    """
    Set emulator bypass mode. Emulator bypass mode will bypass the entire emulator
    for minimum latency.
    """

    code: typing.ClassVar[int] = 454
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, whether the emulator bypass is enabled.

    @dataclass(frozen=True)
    class GetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, whether the emulator bypass is enabled.

    def get(self) -> "Token[GetDataAttr]":
        """Get the status of bypass mode of the impairment emulator.

        :return: the status of bypass mode of the impairment emulator.
        :rtype: M_EMULBYPASS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, on_off: OnOff) -> "Token":
        """Set the bypass mode of the impairment emulator.

        :param on_off: the bypass mode of the impairment emulator.
        :type on_off: OnOff
        """
        return Token(self._connection, build_set_request(self, module=self._module, on_off=on_off))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Disable the bypass mode of the impairment emulator.
    """
    set_on = functools.partialmethod(set, OnOff.ON)
    """Enable the bypass mode of the impairment emulator.
    """



from enum import IntEnum
from typing import Final


NOTHING: Final = 0xFF
WILDCARD: Final = 0xFE
NOVALUE: Final = "N/A"
MAGIC_WORD: Final = b"XENA"
# MAGIC_WORD = 0x58454E41


class CommandType(IntEnum):
    """BXMP request type"""

    COMMAND_CHAIN = 1
    COMMAND_QUERY = 2
    COMMAND_VALUE = 3
    COMMAND_STATUS = 4


class CommandStatus(IntEnum):
    """Status codes in BXMP response message to request type <COMMAND_STATUS = 4>"""

    OK = 0  # OK
    NOCONNECTIONS = 1  # chassis has no available connection slots
    NOTLOGGEDON = 2  # no command can be submitted before logon
    NOTRESERVED = 3  # parameter cannot be set because containing resource not reserved
    NOTWRITABLE = 4  # parameter cannot be set because it is not in a writable state
    NOTREADABLE = 5  # parameter is write-only
    NOTVALID = 6  # operation not valid in current state
    BADHEADER = 7  # invalid magic word
    BADCOMMAND = 8  # invalid command
    BADPARAMETER = 9  # invalid parameter code
    BADMODULE = 10  # invalid module index
    BADPORT = 11  # invalid port index
    BADINDEX = 12  # invalid parameter index
    BADSIZE = 13  # invalid size of data
    BADVALUE = 14  # invalid value of data
    FAILED = 15  # failed to perform operation
    MEMORYFAILURE = 16  # failed to allocate memory
    NOLICPE = 17  # no free PE license
    NOLICPORT = 18  # no free Port license
    NOTSUPPORTED = 19  # feature not supported
    PENDING = 20  # status return will wait until command has been executed
    MODULE_OPERATION_NOT_SUPPORTED_BY_CHASSIS = 21  # module is not supported by chassis - e.g. because multi-image requires x64 OS.
    XLSFAILED = 22  # could not establish connection to Xena License Server
    XLSDENIED = 23  # request for resource rejected by Xena License Server
    XLSINVALID = 24  # Trying to run Valkyrie VE with VulcanVE-300 resource



import functools
from typing_extensions import Self
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
    TypeVar,
    Generic,
)
from xoa_driver.internals.core.commands import enums
from xoa_driver.internals.core.commands import (
    P_RESERVATION,
    P_RESERVEDBY,
    P_RESET,
    P_COMMENT,
    P_INTERFACE,
    P_RECEIVESYNC,
)
from xoa_driver.internals.state_storage import ports_state
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.utils import kind

if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf


PortStateStorage = TypeVar('PortStateStorage', bound="ports_state.PortLocalState")


class BasePort(ABC, Generic[PortStateStorage]):
    """Layout which is relevant to all ports."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self._conn = conn
        self.kind = kind.PortKind(module_id, port_id)
        self.sync_status = P_RECEIVESYNC(conn, module_id, port_id)
        """Port sync status.
        
        :type: P_RECEIVESYNC
        """
        self.interface = P_INTERFACE(conn, module_id, port_id)
        """Physical interface type of the port.
        
        :type: P_INTERFACE
        """
        self.reservation = P_RESERVATION(self._conn, *self.kind)
        """Port reservation action.
        
        :type: P_RESERVATION
        """
        self.reserved_by = P_RESERVEDBY(self._conn, *self.kind)
        """Port reservation status.
        
        :type: P_RESERVEDBY
        """
        self.reset = P_RESET(self._conn, *self.kind)
        """Port reset action.
        
        :type: P_RESET
        """
        self.comment = P_COMMENT(self._conn, *self.kind)
        """Port description.
        
        :type: P_COMMENT
        """

    def __await__(self):
        return self._setup().__await__()

    @abstractmethod
    async def _setup(self) -> Self:
        return self

    def _check_identity(self, request) -> bool:
        validators = (
            self.kind.module_id == request.header.module_index,
            self.kind.port_id == request.header.port_index
        )
        return all(validators)

    def __is_reservation(self, reserved_status: enums.ReservedStatus) -> bool:
        return self.info.reservation == reserved_status

    is_released = functools.partialmethod(__is_reservation, enums.ReservedStatus.RELEASED)
    """
    Check if port is released
    """

    is_reserved_by_me = functools.partialmethod(__is_reservation, enums.ReservedStatus.RESERVED_BY_YOU)
    """
    Check if port is released by me
    """

    is_reserved_by_others = functools.partialmethod(__is_reservation, enums.ReservedStatus.RESERVED_BY_OTHER)
    """
    Check if port is released by others
    """

    @property
    @abstractmethod
    def info(self) -> PortStateStorage:
        """Module info"""
        raise NotImplementedError()

    on_reservation_change = functools.partialmethod(utils.on_event, P_RESERVATION)
    """
    Register a callback to the event that the port's reservation status changes.
    """

    on_receive_sync_change = functools.partialmethod(utils.on_event, P_RECEIVESYNC)
    """
    Register a callback to the event that the port's SYNC status changes.
    """

    on_reserved_by_change = functools.partialmethod(utils.on_event, P_RESERVEDBY)
    """
    Register a callback to the event that the port's reservation ownership changes.
    """

    on_interface_change = functools.partialmethod(utils.on_event, P_INTERFACE)
    """
    Register a callback to the event that the port's physical interface type changes.
    """



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands import (
    P4G_USER_STATE_CURRENT,
    P4G_USER_STATE_TOTAL,
    P4G_USER_STATE_RATE,
)


class GUserState:
    """User state counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.current = P4G_USER_STATE_CURRENT(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_USER_STATE_CURRENT
        """
        self.total = P4G_USER_STATE_TOTAL(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_USER_STATE_TOTAL
        """
        self.rate = P4G_USER_STATE_RATE(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_USER_STATE_RATE
        """



from typing import (
    get_args,
    Type,
    ClassVar,
    TypeVar,
)

T = TypeVar("T")


def add_on(data_class: Type) -> Type:
    class AddOn(data_class):
        size: ClassVar[int] = sum(
            get_args(field_type)[0].size
            for field_type in data_class.__annotations__.values()
        )
        is_add_on: ClassVar[bool] = True
        __qualname__ = data_class.__name__

        @classmethod
        def from_bytes(cls: Type[T], data: bytes) -> "T":
            dic = {}
            pointer = 0
            for field_name, field_type in data_class.__annotations__.items():
                typings = get_args(field_type)[0]
                reading_size = typings.size
                read_bytes = data[pointer: pointer + reading_size]
                pointer += reading_size
                dic[field_name] = typings.from_bytes(read_bytes)

            return cls(**dic)

        def __bytes__(self) -> bytes:
            result = b"".join(
                bytes(getattr(self, field_name))
                for field_name in data_class.__annotations__.keys()
            )
            return result

    AddOn.__name__ = data_class.__name__
    return AddOn



from typing import (
    Protocol,
    List,
    Type,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from .. import kind
    from . import observer


class IIndexType(Protocol):
    def __init__(self, conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "observer.IndicesObserver") -> None: ...  # noqa: E704

    async def delete(self) -> None: ...  # noqa: E704

    @property
    def idx(self) -> int: ...  # noqa: E704

    @classmethod
    async def _fetch(cls, conn, module_id, port_id) -> List[int]: ...  # noqa: E704

    @classmethod
    async def _new(cls, conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "observer.IndicesObserver") -> Type: ...  # noqa: E704



from typing import (
    TYPE_CHECKING,
    List,
    TypeVar,
    Type,
)
from xoa_driver.internals.core.commands import (
    PD_INDICES,
    PD_CREATE,
    PD_DELETE,
    PD_ENABLE,
    PD_SOURCE,
    PD_RANGE,
    PD_SAMPLES,
)

if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from xoa_driver.internals.utils import kind
from xoa_driver.internals.utils.indices import observer as idx_obs
from .base_index import BaseIndex


PD = TypeVar("PD")


class PortDatasetIdx(BaseIndex):
    """L23 Histogram Index Manager"""
    def __init__(self, conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> None:
        super().__init__(conn, kind, observer)
        self.enable = PD_ENABLE(conn, *kind)
        """Enable histogram.

        :type: PD_ENABLE
        """

        self.source = PD_SOURCE(conn, *kind)
        """The source criteria specifying what is counted, and for which packets, by a histogram of a port.

        :type: PD_SOURCE
        """

        self.range = PD_RANGE(conn, *kind)
        """The bucket ranges used for classifying the packets counted by a histogram of a port.
        
        :type: PD_RANGE
        """

        self.samples = PD_SAMPLES(conn, *kind)
        """The current set of counts collected by a histogram for a port.

        :type: PD_SAMPLES
        """

    async def delete(self):
        await PD_DELETE(self._conn, *self.kind).set()
        """Delete a histogram.

        :type: PD_ENABLE
        """

        self._observer.notify(idx_obs.IndexEvents.DEL, self)

    @classmethod
    async def _fetch(cls, conn: "itf.IConnection", module_id: int, port_id: int) -> List[int]:
        resp = await PD_INDICES(conn, module_id, port_id).get()
        return list(resp.histogram_indices)

    @classmethod
    async def _new(cls: Type[PD], conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> PD:
        await PD_CREATE(conn, *kind).set()
        return cls(conn, kind, observer)



from __future__ import annotations
from dataclasses import dataclass, field
import typing as t
from xoa_driver import enums
from xoa_driver.utils import apply
from xoa_driver.internals.hli_v1.ports.port_l23.family_l import FamilyL
from xoa_driver.internals.hli_v1.ports.port_l23.family_l1 import FamilyL1
from xoa_driver.ports import GenericL23Port
from xoa_driver.lli import commands
from xoa_driver.internals.core import interfaces as itf
from xoa_driver.misc import Token
from .tools import (
    get_ctx,
    dictionize_autoneg_status,
    dictionize_lt_status,
    dictionize_txtap_get,
    dictionize_anlt_status,
    dictionize_lt_im_status,
    dictionize_lt_algorithm_status
)
import asyncio

PcsPmaSupported = (FamilyL, FamilyL1)
AutoNegSupported = (FamilyL, FamilyL1)
LinkTrainingSupported = FamilyL


@dataclass
class DoAnlt:
    port: GenericL23Port
    """port object"""
    should_do_an: bool
    """should the port do autoneg?"""
    should_do_lt: bool
    """should the port do link training?"""
    an_allow_loopback: bool
    """should the autoneg allow loopback?"""
    lt_preset0: enums.NRZPreset
    """out-of-sync tap values (preset 0): existing or standard"""
    lt_initial_modulations: dict[str, enums.LinkTrainEncoding]
    """the initial modulations of each serdes"""
    should_lt_interactive: bool
    """should perform link training manually?"""
    lt_algorithm: dict[str, enums.LinkTrainAlgorithm]
    """link training algorithm should be used?"""

    _group: tuple["itf.IConnection", int, int] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        self._group = get_ctx(self.port)

    def __pp_autoneg(self, on: bool) -> Token:
        state = enums.AutoNegMode.ANEG_ON if on else enums.AutoNegMode.ANEG_OFF
        return commands.PP_AUTONEG(*self._group).set(
            state,
            enums.AutoNegTecAbility.DEFAULT_TECH_MODE,
            enums.AutoNegFECOption.NO_FEC,
            enums.AutoNegFECOption.NO_FEC,
            enums.PauseMode.NO_PAUSE,
        )

    def __pp_link_train(
        self,
        mode: enums.LinkTrainingMode,
        nrz_preset: enums.NRZPreset,
        timeout_mode: enums.TimeoutMode,
    ) -> Token:
        return commands.PP_LINKTRAIN(*self._group).set(
            mode=mode,
            pam4_frame_size=enums.PAM4FrameSize.P16K_FRAME,
            nrz_pam4_init_cond=enums.LinkTrainingInitCondition.NO_INIT,
            nrz_preset=nrz_preset,
            timeout_mode=timeout_mode,
        )

    def __pl1_cfg_tmp(
        self, serdes: int, config_type: int, values: list[int]
    ) -> Token:
        return commands.PL1_CFG_TMP(*self._group, serdes, config_type).set(
            values=values
        )

    def __select_modes(self) -> tuple[enums.LinkTrainingMode, enums.TimeoutMode]:
        if self.should_do_an == True and self.should_lt_interactive == False:
            lt_mode = enums.LinkTrainingMode.START_AFTER_AUTONEG
            timeout_mode = enums.TimeoutMode.DEFAULT
        elif self.should_do_an == True and self.should_lt_interactive == True:
            lt_mode = enums.LinkTrainingMode.INTERACTIVE
            timeout_mode = enums.TimeoutMode.DISABLED
        elif self.should_do_an == False and self.should_lt_interactive == True:
            lt_mode = enums.LinkTrainingMode.INTERACTIVE
            timeout_mode = enums.TimeoutMode.DISABLED
        else:
            lt_mode = enums.LinkTrainingMode.STANDALONE
            timeout_mode = enums.TimeoutMode.DEFAULT
        return lt_mode, timeout_mode

    def __builder__(self) -> t.Generator[Token, None, None]:
        """Defining commands sequence"""
        nrz_preset = self.lt_preset0
        # # Set autoneg timeout
        yield self.__pp_link_train(
            enums.LinkTrainingMode.DISABLED,
            enums.NRZPreset.NRZ_NO_PRESET,
            enums.TimeoutMode.DEFAULT,
        )

        # # Set autoneg allow-loopback
        yield self.__pl1_cfg_tmp(
            0, enums.Layer1ConfigType.AN_LOOPBACK, [int(self.an_allow_loopback)]
        )

        # yield self.__pp_autoneg(self.should_do_an and not self.should_do_lt)
        if (not self.should_do_an) or self.should_do_lt:
            # Disable autoneg
            yield self.__pp_autoneg(False)

        if self.should_do_lt:
            for serdes_str, algorithm in self.lt_algorithm.items():
                # # Set the link train algorithm
                yield self.__pl1_cfg_tmp(
                    int(serdes_str), enums.Layer1ConfigType.LT_TRAINING_ALGORITHM, [algorithm.value]
                )

            for serdes_str, im in self.lt_initial_modulations.items():
                yield self.__pl1_cfg_tmp(
                    int(serdes_str), enums.Layer1ConfigType.LT_INITIAL_MODULATION, [im.value]
                )

            lt_mode, timeout_mode = self.__select_modes()
            yield self.__pp_link_train(
                enums.LinkTrainingMode.DISABLED, nrz_preset, timeout_mode
            )
            yield self.__pp_link_train(lt_mode, nrz_preset, timeout_mode)

        if self.should_do_an:
            yield self.__pp_autoneg(True)

    async def run(self) -> None:
        """Start anlt execution"""
        await apply(*self.__builder__())


async def anlt_start(
    port: GenericL23Port,
    should_do_an: bool,
    should_do_lt: bool,
    an_allow_loopback: bool,
    lt_preset0: enums.NRZPreset,
    lt_initial_modulations: dict[str, enums.LinkTrainEncoding],
    should_lt_interactive: bool,
    lt_algorithm: dict[str, enums.LinkTrainAlgorithm],
) -> None:
    """
    .. versionadded:: 1.1

    Start ANLT on a port

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param should_do_an: should the port do autoneg?
    :type should_do_an: bool
    :param should_do_lt: should the port do link training?
    :type should_do_lt: bool
    :param an_allow_loopback: should the autoneg allow loopback?
    :type an_allow_loopback: bool
    :param lt_preset0: out-of-sync tap values (preset 0): existing or standard
    :type lt_preset0: enums.NRZPreset
    :param lt_initial_modulations: the initial modulations of each serdes
    :type lt_initial_modulations: typing.Dict[str, enums.LinkTrainEncoding]
    :param should_lt_interactive: should perform link training manually?
    :type should_lt_interactive: bool
    :param lt_algorithm: Link training algorithm to use
    :type lt_algorithm: Dict[str, enums.LinkTrainAlgorithm]
    """

    anlt = DoAnlt(
        port,
        should_do_an,
        should_do_lt,
        an_allow_loopback,
        lt_preset0,
        lt_initial_modulations,
        should_lt_interactive,
        lt_algorithm,
    )
    await anlt.run()


async def autoneg_status(port: GenericL23Port) -> dict[str, t.Any]:
    """
    .. versionadded:: 1.1

    Get the auto-negotiation status

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :return:
    :rtype: typing.Dict[str, typing.Any]
    """
    conn, mid, pid = get_ctx(port)
    loopback, auto_neg_info, status = await apply(
        commands.PL1_CFG_TMP(
            conn, mid, pid, 0, enums.Layer1ConfigType.AN_LOOPBACK
        ).get(),
        commands.PL1_AUTONEGINFO(conn, mid, pid, 0).get(),
        commands.PP_AUTONEGSTATUS(conn, mid, pid).get(),
    )
    return dictionize_autoneg_status(loopback, auto_neg_info, status)


LinkTrainType = t.Union[
    enums.LinkTrainCoeffs,
    enums.LinkTrainPresets,
    enums.LinkTrainEncoding,
    enums.LinkTrainAnnounce,
]


async def __lt_coeff(
    port: GenericL23Port,
    serdes: int,
    arg: LinkTrainType,
    *,
    cmd: enums.LinkTrainCmd,
) -> enums.LinkTrainCmdResults:
    conn, mid, pid = get_ctx(port)
    cmd_ = commands.PL1_LINKTRAIN_CMD(conn, mid, pid, serdes)
    await cmd_.set(cmd=cmd, arg=arg.value)
    for _ in range(1000):
        resp = await cmd_.get()
        status = resp.result
        if (resp.flags & enums.LinkTrainCmdFlags.DONE.value):
            return enums.LinkTrainCmdResults(status)
        await asyncio.sleep(0.01)
    return enums.LinkTrainCmdResults.UNKNOWN


async def lt_coeff_inc(
    port: GenericL23Port, serdes: int, emphasis: enums.LinkTrainCoeffs
) -> enums.LinkTrainCmdResults:
    """
    .. versionadded:: 1.1
    
    Ask the remote port to increase coeff of the specified serdes.

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param serdes: the serdes index, starting from 0
    :type serdes: int
    :param emphasis: the emphasis to increase
    :type emphasis: enums.LinkTrainCoeffs
    :return:
    :rtype: None
    """
    return await __lt_coeff(port, serdes, emphasis, cmd=enums.LinkTrainCmd.CMD_INC)


async def lt_coeff_dec(
    port: GenericL23Port, serdes: int, emphasis: enums.LinkTrainCoeffs
) -> enums.LinkTrainCmdResults:
    """
    .. versionadded:: 1.1
    
    Ask the remote port to decrease coeff of the specified serdes.

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param serdes: the serdes index, starting from 0
    :type serdes: int
    :param emphasis: the emphasis to decrease
    :type emphasis: enums.LinkTrainCoeffs
    :return:
    :rtype: None
    """
    return await __lt_coeff(port, serdes, emphasis, cmd=enums.LinkTrainCmd.CMD_DEC)


async def lt_preset(
    port: GenericL23Port, serdes: int, preset: enums.LinkTrainPresets
) -> enums.LinkTrainCmdResults:
    """
    .. versionadded:: 1.1
    
    Ask the remote port to use the preset of the specified serdes.

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param serdes: the serdes index, starting from 0
    :type serdes: int
    :param preset: preset index to select for the serdes, 0,1,2,3,4,
    :type preset: enums.LinkTrainPresets
    :return:
    :rtype: None
    """
    return await __lt_coeff(port, serdes, preset, cmd=enums.LinkTrainCmd.CMD_PRESET)


async def lt_encoding(
    port: GenericL23Port, serdes: int, encoding: enums.LinkTrainEncoding
) -> enums.LinkTrainCmdResults:
    """
    .. versionadded:: 1.1
    
    Ask the remote port to use the encoding of the specified serdes.

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param serdes: the serdes index, starting from 0
    :type serdes: int
    :param encoding: link training encoding
    :type encoding: enums.LinkTrainCoeffs
    :return:
    :rtype: None
    """
    return await __lt_coeff(port, serdes, encoding, cmd=enums.LinkTrainCmd.CMD_ENCODING)


async def lt_trained(port: GenericL23Port, serdes: int) -> enums.LinkTrainCmdResults:
    """
    .. versionadded:: 1.1
    
    Tell the remote port that the current serdes is trained.

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param serdes: the serdes index, starting from 0
    :type serdes: int
    :return:
    :rtype: None
    """
    return await __lt_coeff(
        port,
        serdes,
        arg=enums.LinkTrainAnnounce.TRAINED,
        cmd=enums.LinkTrainCmd.CMD_LOCAL_TRAINED,
    )


async def lt_status(port: GenericL23Port, serdes: int) -> dict[str, t.Any]:
    """
    .. versionadded:: 1.1
    
    Show the link training status.

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param serdes: the serdes index, starting from 0
    :type serdes: int
    :return: LT status of the serdes
    :rtype: typing.Dict[str, typing.Any]
    """
    conn, mid, pid = get_ctx(port)
    status, info, ltconf, cfg = await apply(
        commands.PP_LINKTRAINSTATUS(conn, mid, pid, serdes).get(),
        commands.PL1_LINKTRAININFO(conn, mid, pid, serdes, 0).get(),
        commands.PP_LINKTRAIN(conn, mid, pid).get(),
        commands.PL1_CFG_TMP(
            conn, mid, pid, serdes, enums.Layer1ConfigType.LT_INITIAL_MODULATION
        ).get(),
    )
    total_bit_count = (info.prbs_total_bits_high << 32) + info.prbs_total_bits_low
    total_error_bit_count = (
        info.prbs_total_error_bits_high << 32
    ) + info.prbs_total_error_bits_low
    ber = (
        total_error_bit_count / total_bit_count if total_bit_count > 0 else float("nan")
    )
    return dictionize_lt_status(
        status, info, ltconf, cfg, ber, total_bit_count, total_error_bit_count
    )


async def txtap_get(port: GenericL23Port, serdes: int) -> dict[str, int]:
    """
    .. versionadded:: 1.1
    
    Get the tap value of the local TX tap.

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param serdes: the serdes index, starting from 0
    :type serdes: int
    :return: tap values of the serdes
    :rtype: typing.Dict[str, int]
    """
    conn, mid, pid = get_ctx(port)
    r = await commands.PP_PHYTXEQ(conn, mid, pid, serdes).get()
    return dictionize_txtap_get(r)


async def txtap_set(
    port: GenericL23Port,
    serdes: int,
    pre3: int,
    pre2: int,
    pre: int,
    main: int,
    post1: int,
) -> None:
    """
    .. versionadded:: 1.1
    
    Set the tap value of the local TX tap.

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param serdes: the serdes index, starting from 0
    :type serdes: int
    :param pre3: pre3 value
    :type pre3: int
    :param pre2: pre2 value
    :type pre2: int
    :param pre: pre value
    :type pre: int
    :param main: main value
    :type main: int
    :param post1: post1 value
    :type post1: int
    :return:
    :rtype: None
    """
    conn, mid, pid = get_ctx(port)
    cmd_ = commands.PP_PHYTXEQ(conn, mid, pid, serdes)
    await cmd_.set(
        pre1=pre,
        main=main,
        post1=post1,
        pre2=pre2,
        post2=pre3,
        post3=0,
    )


async def anlt_link_recovery(port: GenericL23Port, enable: bool) -> None:
    """
    .. versionadded:: 1.1
    
    Should xenaserver automatically do link recovery when detecting down signal.

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param enable: should the port automatically do link recovery when link is down.
    :type enable: bool
    :return:
    :rtype:  None
    """
    conn, mid, pid = get_ctx(port)
    cmd_ = commands.PL1_CFG_TMP(
        conn, mid, pid, 0, enums.Layer1ConfigType.ANLT_INTERACTIVE
    )
    await cmd_.set(values=[int(enable)])


async def anlt_status(port: GenericL23Port) -> dict[str, t.Any]:
    """
    .. versionadded:: 1.1
    
    Get the overview of ANLT status

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :return: AN/LT status of the port
    :rtype: typing.Dict[str, typing.Any]
    """

    # if not isinstance(port, LinkTrainingSupported):
    #     raise NotSupportLinkTrainError(port)
    conn, mid, pid = get_ctx(port)
    r = await apply(
        commands.PL1_CFG_TMP(
            conn, mid, pid, 0, enums.Layer1ConfigType.ANLT_INTERACTIVE
        ).get(),
        commands.PP_AUTONEGSTATUS(conn, mid, pid).get(),
        commands.PP_LINKTRAIN(conn, mid, pid).get(),
        commands.P_CAPABILITIES(conn, mid, pid).get(),
        commands.PL1_CFG_TMP(conn, mid, pid, 0, enums.Layer1ConfigType.AN_LOOPBACK).get(),
    )
    link_recovery, autoneg, linktrain, capabilities, allow_loopback= r

    return dictionize_anlt_status(link_recovery, autoneg, linktrain, capabilities, allow_loopback)


async def anlt_log(port: GenericL23Port) -> str:
    """
    .. versionadded:: 1.1
    
    Get the anlt log messages

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :return: AN/LT protocol log traces of the port
    :rtype: str
    """
    conn, mid, pid = get_ctx(port)
    log = await commands.PL1_LOG(conn, mid, pid).get()
    return log.log_string


async def anlt_stop(port: GenericL23Port) -> None:
    """
    .. versionadded:: 1.3

    Stop AN & LT

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    """
    
    anlt = DoAnlt(
        port=port,
        should_do_an=False,
        should_do_lt=False,
        an_allow_loopback=False,
        lt_preset0=enums.NRZPreset.NRZ_NO_PRESET,
        lt_initial_modulations={},
        should_lt_interactive=False,
        lt_algorithm={}
    )
    await anlt.run()


async def txtap_autotune(
    port: GenericL23Port,
    serdes: int,
) -> None:
    """
    .. versionadded:: 1.3
    
    Auto tune the tap value of the local TX tap.

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param serdes: the serdes index, starting from 0
    :type serdes: int
    :return:
    :rtype: None
    """
    conn, mid, pid = get_ctx(port)
    phy_autotune = commands.PP_PHYAUTOTUNE(conn, mid, pid, serdes)
    await phy_autotune.set(on_off=enums.OnOff.OFF)
    await phy_autotune.set(on_off=enums.OnOff.ON)


async def lt_im_status(port: GenericL23Port) -> dict[str, t.Any]:
    """
    .. versionadded:: 1.3
    
    Get LT initial modulation config

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :return: LT initial modulation configuration of the port
    :rtype: typing.Dict[str, typing.Any]
    """

    # if not isinstance(port, LinkTrainingSupported):
    #     raise NotSupportLinkTrainError(port)
    conn, mid, pid = get_ctx(port)
    capabilities = await commands.P_CAPABILITIES(conn, mid, pid).get()
    initial_mods = {}
    for i in range(0, capabilities.serdes_count):
        im = await commands.PL1_CFG_TMP(conn, mid, pid, i, enums.Layer1ConfigType.LT_INITIAL_MODULATION).get()
        initial_mods[str(i)] = enums.LinkTrainEncoding(im.values[0]).name

    return dictionize_lt_im_status(capabilities, initial_mods)


async def lt_algorithm_status(port: GenericL23Port) -> dict[str, t.Any]:
    """
    .. versionadded:: 1.3
    
    Get LT initial modulation config

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :return: LT initial modulation configuration of the port
    :rtype: typing.Dict[str, typing.Any]
    """

    # if not isinstance(port, LinkTrainingSupported):
    #     raise NotSupportLinkTrainError(port)
    conn, mid, pid = get_ctx(port)
    capabilities = await commands.P_CAPABILITIES(conn, mid, pid).get()
    algorithms = {}
    for i in range(0, capabilities.serdes_count):
        alg = await commands.PL1_CFG_TMP(conn, mid, pid, i, enums.Layer1ConfigType.LT_TRAINING_ALGORITHM).get()
        algorithms[str(i)] = enums.LinkTrainAlgorithm(alg.values[0]).name

    return dictionize_lt_algorithm_status(capabilities, algorithms)


async def anlt_strict(port: GenericL23Port, enable: bool) -> None:
    """
    .. versionadded:: 1.3
    
    Should ANLT strict mode be enabled

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param enable: should ANLT strict mode be enabled
    :type enable: bool
    :return:
    :rtype:  None
    """
    conn, mid, pid = get_ctx(port)
    capabilities = await commands.P_CAPABILITIES(conn, mid, pid).get()
    for i in range(0, capabilities.serdes_count):
        await commands.PL1_CFG_TMP(conn, mid, pid, i, enums.Layer1ConfigType.ANLT_STRICT_MODE).set(values=[int(enable)])


async def anlt_log_control(port: GenericL23Port, types: t.List[enums.AnLtLogControl]) -> None:
    """
    .. versionadded:: 1.3
    
    Control what should be logged for ANLT by xenaserver

    :param port: the port object
    :type port: :class:`~xoa_driver.ports.GenericL23Port`
    :param types: control what should be logged for ANLT by xenaserver
    :type types: t.List[enums.AnLtLogControl]
    :return:
    :rtype:  None
    """
    conn, mid, pid = get_ctx(port)
    capabilities = await commands.P_CAPABILITIES(conn, mid, pid).get()
    type = 0
    for _type in types:
        type |= _type.value
    for i in range(0, capabilities.serdes_count):
        await commands.PL1_CFG_TMP(conn, mid, pid, i, enums.Layer1ConfigType.ANLT_LOG_CONTROL).set(values=[int(type)])


__all__ = (
    "anlt_link_recovery",
    "anlt_log",
    "anlt_start",
    "anlt_status",
    "anlt_stop",
    "autoneg_status",
    "lt_coeff_inc",
    "lt_coeff_dec",
    "lt_encoding",
    "lt_preset",
    "lt_status",
    "lt_trained",
    "txtap_get",
    "txtap_set",
    "txtap_autotune",
    "lt_im_status",
    "lt_algorithm_status",
    "anlt_strict",
    "anlt_log_control",
)



#: L23 Port Filter Commands

from dataclasses import dataclass
import typing
import functools

from ..protocol.command_builders import (
    build_get_request,
    build_set_request
)
from .. import interfaces
from ..transporter.token import Token
from ..protocol.fields import data_types as xt
from ..protocol.fields.field import XmpField
from ..registry import register_command
from .enums import *  # noqa: F403


@register_command
@dataclass
class PF_INDICES:
    """
    The full list of which filters are defined for a port. These are the sub-index
    values that are used for the parameters defining the compound conditions on the
    match/length terms operating on the packets received for the port. Setting the
    value of this parameter creates a new empty filter for each value that is not
    already in use, and deletes each filter that is not mentioned in the list. The
    same can be accomplished one-filter-at-a-time using the PF_CREATE and PF_DELETE
    commands.
    """

    code: typing.ClassVar[int] = 211
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    @dataclass(frozen=True)
    class SetDataAttr:
        filter_xindices: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)  # list of integers, the list of indices of filters on a port.

    @dataclass(frozen=True)
    class GetDataAttr:
        filter_xindices: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)  # list of integers, the list of indices of filters on a port.

    def get(self) -> "Token[GetDataAttr]":
        """Get the full list of which filters are defined for a port.

        :return: the list of indices of filters on a port
        :rtype: PF_INDICES.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, filter_xindices: typing.List[int]) -> "Token":
        """Create a new empty filter for each index value that is not already in use, and deletes each filter that is not mentioned in the list.

        :param filter_xindices: the list of indices of filters to be created on a port.
        :type filter_xindices: List[int]
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, filter_xindices=filter_xindices))


@register_command
@dataclass
class PF_CREATE:
    """
    Creates an empty filter definition with the specified sub-index value.
    """

    code: typing.ClassVar[int] = 212
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _filter_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        pass

    def set(self) -> "Token":
        """Creates an empty filter definition with the specified sub-index value.
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._filter_xindex],
            ),
        )


@register_command
@dataclass
class PF_DELETE:
    """
    Deletes the filter definition with the specified sub-index value.
    """

    code: typing.ClassVar[int] = 213
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _filter_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        pass

    def set(self) -> "Token":
        """Delete the filter definition with the specified sub-index value.
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._filter_xindex],
            ),
        )


@register_command
@dataclass
class PF_ENABLE:
    """
    Whether a filter is currently active on a port. While a filter is enabled its
    condition cannot be changed, nor can any match term or length terms used by it.
    """

    code: typing.ClassVar[int] = 214
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _filter_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, whether the filter is enabled.

    @dataclass(frozen=True)
    class GetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, whether the filter is enabled.

    def get(self) -> "Token[GetDataAttr]":
        """Get whether a filter is currently active on the port.

        :return: whether the filter is enabled
        :rtype: PF_ENABLE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._filter_xindex]))

    def set(self, on_off: OnOff) -> "Token":
        """Set whether a filter is currently active on the port.

        :param on_off: whether the filter is enabled
        :type on_off: OnOff
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._filter_xindex], on_off=on_off))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Disable a filter on a port.
    """
    set_on = functools.partialmethod(set, OnOff.ON)
    """Enable a filter on a port.
    """


@register_command
@dataclass
class PF_COMMENT:
    """
    The description of a filter.
    """

    code: typing.ClassVar[int] = 215
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _filter_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        comment: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the description of the filter.

    @dataclass(frozen=True)
    class GetDataAttr:
        comment: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the description of the filter.

    def get(self) -> "Token[GetDataAttr]":
        """Get the description of a filter.

        :return: the description of the filter
        :rtype: PF_COMMENT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._filter_xindex]))

    def set(self, comment: str) -> "Token":
        """Set the description of a filter.

        :param comment: the description of the filter.
        :type comment: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._filter_xindex], comment=comment))


@register_command
@dataclass
class PF_CONDITION:
    """
        The boolean condition on the terms specifying when the filter is satisfied. The condition uses a canonical and-or-not expression on the match terms and length terms.
        The condition is specified using a number of compound terms, each encoded as an integer value specifying an arbitrary set of the match terms
        and length terms defined for the port. Each match or length term has a specific power-of-two value, and the set is encoded as the sum of the values for the contained terms:

        Value for match term ``[match_term_xindex] = 2^match_term_xindex``

        Value for length term ``[length_term_xindex] = 2^(length_term_xindex+16)``

        A compound term is true if all the match terms and length terms contained in it are true. This supports the and-part of the condition.
        If some compound term is satisfied, the condition as a whole is true.

        This is the or-part of the condition. The first few compound terms at the even positions (second, fourth, ...) are inverted,
        and all the contained match terms and length terms must be false at the same time that the those of the preceding compound term are true.
        This is the not-part of the condition.

        At the top level, a condition is a bunch of things or-ed together.

        ``<filter-condition> = <or-expr>``

        Two of the or-operands are *general*, two are 'simple'.

        ``<or-expr> =  <general-and-expr>  or  <general-and-expr>  or  <simple-and-expr>  or  <simple-and-expr>``

        A 'general' and-expression can include negated terms.

        ``<general-and-expr>  =  <term>  and  <term>  and ... and  not <term>  and ... and  not <term>``

        A 'simple' and-expression can only have non-negated terms.

        ``<simple-and-expr>   =  <term>  and  <term>  and ... and <term>``

        ``<term>              =  <match-term>``

        ``<term>              =  <length-term>``

        In practice, the simplest way to generate these encodings is to use the ValkyrieManager,
        which supports Boolean expressions using the operators ``&, |, and ~``, and simply query the chassis for the resulting script-level definition.

    """

    code: typing.ClassVar[int] = 216
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _filter_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        and_expression_0: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match terms AND length terms.
        and_not_expression_0: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match NOT terms AND length NOT terms.
        and_expression_1: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match terms AND length terms.
        and_not_expression_1: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match NOT terms AND length NOT terms.
        and_expression_2: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match terms AND length terms.
        and_expression_3: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match terms AND length terms.

    @dataclass(frozen=True)
    class GetDataAttr:
        and_expression_0: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match terms AND length terms.
        and_not_expression_0: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match NOT terms AND length NOT terms.
        and_expression_1: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match terms AND length terms.
        and_not_expression_1: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match NOT terms AND length NOT terms.
        and_expression_2: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match terms AND length terms.
        and_expression_3: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # unsigned integer, encoding a compound term which is a set of the match terms AND length terms.

    def get(self) -> "Token[GetDataAttr]":
        """Get the condition on the terms specifying when the filter is satisfied.

        :return: and_expression_0, and_not_expression_0, and_expression_1, and_not_expression_1, and_expression_2, and and_expression_3.

        :rtype: ~PF_CONDITION.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._filter_xindex]))

    def set(self, and_expression_0: int, and_not_expression_0: int, and_expression_1: int, and_not_expression_1: int, and_expression_2: int, and_expression_3: int) -> "Token":
        """Set the condition on the terms specifying when the filter is satisfied.

        :param and_expression_0: encoding a compound term which is a set of the match terms AND length terms.
        :type and_expression_0: int
        :param and_not_expression_0: encoding a compound term which is a set of the match NOT terms AND length NOT terms.
        :type and_not_expression_0: int
        :param and_expression_1: encoding a compound term which is a set of the match terms AND length terms.
        :type and_expression_1: int
        :param and_not_expression_1: encoding a compound term which is a set of the match NOT terms AND length NOT terms.
        :type and_not_expression_1: int
        :param and_expression_2: encoding a compound term which is a set of the match terms AND length terms.
        :type and_expression_2: int
        :param and_expression_3: encoding a compound term which is a set of the match terms AND length terms.
        :type and_expression_3: int

        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._filter_xindex],
                and_expression_0=and_expression_0,
                and_not_expression_0=and_not_expression_0,
                and_expression_1=and_expression_1,
                and_not_expression_1=and_not_expression_1,
                and_expression_2=and_expression_2,
                and_expression_3=and_expression_3
            )
        )


@register_command
@dataclass
class PF_STRING:
    """
    The string representation of a filter.
    """

    code: typing.ClassVar[int] = 217
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _filter_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        string_name: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the string representation of the filter.

    @dataclass(frozen=True)
    class GetDataAttr:
        string_name: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the string representation of the filter.

    def get(self) -> "Token[GetDataAttr]":
        """Get the string representation of a filter.

        :return: the string representation of a filter
        :rtype: ~PF_STRING.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._filter_xindex]))

    def set(self, string_name: str) -> "Token":
        """Set the string representation of a filter.

        :param string_name: the string representation of the filter
        :type string_name: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._filter_xindex], string_name=string_name))



#: Other types which are used by ports or as a parameter to attribute set method.


from xoa_driver.internals.core.transporter.token import Token
from xoa_driver.internals.core.commands.subtypes import (
    ArpChunk,
    NdpChunk,
)

# indices types
from xoa_driver.internals.hli_v2.indices.connection_group.cg import ConnectionGroupIdx as ConnectionGroup
from xoa_driver.internals.hli_v2.indices.filter.base_filter import BaseFilterIdx as BasePortFilter
from xoa_driver.internals.hli_v2.indices.filter.genuine_filter import GenuineFilterIdx as GenuinePortFilter
from xoa_driver.internals.hli_v2.indices.length_term import LengthTermIdx as LengthTerm
from xoa_driver.internals.hli_v2.indices.match_term import MatchTermIdx as MatchTerm
from xoa_driver.internals.hli_v2.indices.port_dataset import PortDatasetIdx as PortDataset
from xoa_driver.internals.hli_v2.indices.streams.base_stream import BaseStreamIdx as BaseStream
from xoa_driver.internals.hli_v2.indices.streams.genuine_stream import GenuineStreamIdx as GenuineStream
from xoa_driver.internals.hli_v2.ports.port_l23.chimera.port_emulation import CFlow as ImpairmentFlow
from xoa_driver.internals.hli_v2.ports.port_l23.chimera.filter_definition.general import ModeBasic as BasicImpairmentFlowFilter
from xoa_driver.internals.hli_v2.ports.port_l23.chimera.filter_definition.general import ModeExtended as ExtendedImpairmentFlowFilter


__all__ = (
    "Token",
    "ArpChunk",
    "NdpChunk",
    "ConnectionGroup",
    "BasePortFilter",
    "GenuinePortFilter",
    "LengthTerm",
    "MatchTerm",
    "PortDataset",
    "BaseStream",
    "GenuineStream",
    "ImpairmentFlow",
    "BasicImpairmentFlowFilter",
    "ExtendedImpairmentFlowFilter",
)



import asyncio
from typing import (
    List,
    Type,
    Generic,
    TypeVar,
    TYPE_CHECKING,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from ._interfaces import IIndexType

from . import observer
from .. import kind


IT = TypeVar("IT", bound="IIndexType")


class IndexManager(Generic[IT]):
    __slots__ = ("_conn", "_idx_type", "_module_id", "_port_id", "_indices", "_lock", "_observer")

    def __init__(self, conn: "itf.IConnection", idx_type: Type[IT], module_id: int, port_id: int) -> None:
        self._conn = conn
        self._idx_type = idx_type
        self._module_id = module_id
        self._port_id = port_id
        self._indices: List[IT] = []
        self._lock = asyncio.Lock()
        self._observer: "observer.IndicesObserver" = observer.IndicesObserver()
        self._observer.subscribe(
            observer.IndexEvents.DEL,
            self.__remove_from_slot
        )

    async def server_sync(self) -> None:
        """Sync the indices with xenaserver"""
        self._indices.clear()
        idxs: List[int] = await self._idx_type._fetch(self._conn, self._module_id, self._port_id)
        for idx_id in idxs:
            index_kind = kind.IndicesKind(
                self._module_id,
                self._port_id,
                idx_id
            )
            idx_instance = self._idx_type(self._conn, index_kind, self._observer)
            self._indices.append(idx_instance)

    def __str__(self) -> str:
        return f"Iterable[{self._idx_type.__name__}]({self._indices!s})"

    def __len__(self) -> int:
        """Return the number of existing indices"""
        return len(self._indices)

    def __iter__(self):
        return iter(self._indices)

    def obtain(self, key: int):
        return self._indices[key]

    def obtain_multiple(self, *keys: int):
        """Obtain multiple resources as a tuple of indices"""
        return tuple(self._indices[k] for k in keys)

    def __detect_empty_idx_slot(self) -> int:
        if len(self._indices) == 0:
            return 0
        existing_indices = [i.idx for i in self._indices]
        empties = [
            ele
            for ele in range(max(existing_indices) + 1)
            if ele not in existing_indices
        ]
        return min(empties) if empties else len(existing_indices)

    def __remove_from_slot(self, index_inst: Type) -> None:
        # throws ValueError if element is not exists in list of indices
        self._indices.remove(index_inst)

    async def create(self):
        async with self._lock:
            index_kind = kind.IndicesKind(
                self._module_id,
                self._port_id,
                self.__detect_empty_idx_slot()
            )
            index_inst: IT = await self._idx_type._new(self._conn, index_kind, self._observer)
            assert index_inst, f"Failed to create Index: {len(self)}"
            self._indices.append(index_inst)
            return index_inst

    async def remove(self, position_idx: int) -> None:
        """Remove an index from the port"""
        await self._indices[position_idx].delete()



from typing import (
    TYPE_CHECKING,
    Dict,
    Type,
    Union
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands.enums import (
    FilterType,
    ProtocolOption,
    FilterMode,
)
from xoa_driver.internals.core.commands import (
    PEF_INIT,
    PEF_APPLY,
    PEF_ENABLE,
    PEF_PROTOCOL,
    PEF_MODE,
)
from . import general


class ModeExtendedS(general.ModeExtended):
    async def use_segments(self, *segments: ProtocolOption) -> None:
        segments_payload = [
            ProtocolOption.ETHERNET,
            *segments
        ]
        await PEF_PROTOCOL(
            self._conn,
            self._module_id,
            self._port_id,
            self._flow_index,
            self._filter_type
        ).set(segments_payload)


class FilterDefinitionShadow:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self._conn = conn
        self._module_id = module_id
        self._port_id = port_id
        self._flow_index = flow_index
        self._filter_type = FilterType.SHADOW

        self.initiating = PEF_INIT(conn, module_id, port_id, flow_index)
        """Preparing filter definition.
        
        :type: PEF_INIT
        """
        self.apply = PEF_APPLY(conn, module_id, port_id, flow_index)
        """Apply filter definition.
        
        :type: PEF_APPLY
        """
        self.enable = PEF_ENABLE(conn, module_id, port_id, flow_index, self._filter_type)
        """Enabling the filter.
        
        :type: PEF_ENABLE
        """
        # self.mode = PEF_MODE(conn, module_id, port_id, flow_index, self._filter_type)
        # """Filter mode.
        # 
        # :type: PEF_MODE`
        # """

    async def get_mode(self) -> Union[general.ModeBasic, ModeExtendedS]:
        mode_types: Dict[FilterMode, Type[Union[general.ModeBasic, ModeExtendedS]]] = {
            FilterMode.BASIC: general.ModeBasic,
            FilterMode.EXTENDED: ModeExtendedS,
        }
        return await general.get_mode(
            self._conn,
            self._module_id,
            self._port_id,
            self._flow_index,
            self._filter_type,
            mode_types
        )

    async def use_basic_mode(self) -> None:
        await PEF_MODE(
            self._conn,
            self._module_id,
            self._port_id,
            self._flow_index,
            self._filter_type,
        ).set_basic()

    async def use_extended_mode(self) -> None:
        await PEF_MODE(
            self._conn,
            self._module_id,
            self._port_id,
            self._flow_index,
            self._filter_type,
        ).set_extended()



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands import (
    P4G_CLEAR_POST_STAT,
    P4G_TCP_RTT_VALUE,
    P4G_TCP_STATE_CURRENT,
    P4G_TCP_STATE_TOTAL,
    P4G_TCP_STATE_RATE,
    P4G_TCP_RX_PAYLOAD_COUNTERS,
    P4G_TCP_TX_PAYLOAD_COUNTERS,
    P4G_TCP_RETRANSMIT_COUNTERS,
    P4G_TCP_ERROR_COUNTERS,
    P4G_TCP_MSS_TYPE,
    P4G_TCP_MSS_MINMAX,
    P4G_TCP_MSS_VALUE,
    P4G_TCP_WINDOW_SIZE,
    P4G_TCP_DUP_THRES,
    P4G_TCP_SYN_RTO,
    P4G_TCP_RTO,
    P4G_TCP_CONGESTION_MODE,
    P4G_TCP_WINDOW_SCALING,
    P4G_TCP_RTO_MINMAX,
    P4G_TCP_RTO_PROLONGED_MODE,
    P4G_TCP_ICWND_CALC_METHOD,
    P4G_TCP_ISSTHRESH,
    P4G_TCP_ACK_FREQUENCY,
    P4G_TCP_ACK_TIMEOUT,
    P4G_TCP_ESTABLISH_HIST,
    P4G_TCP_CLOSE_HIST,
    P4G_TCP_RX_TOTAL_BYTES_HIST,
    P4G_TCP_RX_GOOD_BYTES_HIST,
    P4G_TCP_TX_TOTAL_BYTES_HIST,
    P4G_TCP_TX_GOOD_BYTES_HIST,
    P4G_TCP_RX_PACKET_COUNTERS,
    P4G_TCP_TX_PACKET_COUNTERS,
)


class GAckTcp:
    """TCP ACK configuration"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.duplicate_thresholds = P4G_TCP_DUP_THRES(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_DUP_THRES
        """
        self.frequency = P4G_TCP_ACK_FREQUENCY(conn, module_id, port_id, group_idx)  # TODO: probably need a better name to be closer to cmd description
        """
        
        :type: P4G_TCP_ACK_FREQUENCY
        """
        self.timeout = P4G_TCP_ACK_TIMEOUT(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_ACK_TIMEOUT
        """


class GRetransmissionTimeoutTcp:
    """TCP Retransmission Timeout configuration"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.syn_value = P4G_TCP_SYN_RTO(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_SYN_RTO
        """
        self.value = P4G_TCP_RTO(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_RTO
        """
        self.range_limits = P4G_TCP_RTO_MINMAX(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_RTO_MINMAX
        """
        self.prolonged_mode = P4G_TCP_RTO_PROLONGED_MODE(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_RTO_PROLONGED_MODE
        """


class GStateCountersTcp:
    """TCP State counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.current = P4G_TCP_STATE_CURRENT(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_STATE_CURRENT
        """
        self.total = P4G_TCP_STATE_TOTAL(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_STATE_TOTAL
        """
        self.rate = P4G_TCP_STATE_RATE(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_STATE_RATE
        """


class GMaxSegmentSize:
    """TCP Max Segment Size Configuration"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.type = P4G_TCP_MSS_TYPE(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_MSS_TYPE
        """
        self.range_limits = P4G_TCP_MSS_MINMAX(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_MSS_MINMAX
        """
        self.fixed_value = P4G_TCP_MSS_VALUE(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_MSS_VALUE
        """


class GPacketCountersTcp:
    """TCP Packet Counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.rx = P4G_TCP_RX_PACKET_COUNTERS(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_RX_PACKET_COUNTERS
        """
        self.tx = P4G_TCP_TX_PACKET_COUNTERS(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_TX_PACKET_COUNTERS
        """


class GPayloadCountersTcp:
    """TCP Payload Counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.rx = P4G_TCP_RX_PAYLOAD_COUNTERS(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_RX_PAYLOAD_COUNTERS
        """
        self.tx = P4G_TCP_TX_PAYLOAD_COUNTERS(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_TX_PAYLOAD_COUNTERS
        """


class GCountersTcp:
    """TCP counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.retransmission = P4G_TCP_RETRANSMIT_COUNTERS(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_RETRANSMIT_COUNTERS
        """
        self.error = P4G_TCP_ERROR_COUNTERS(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_ERROR_COUNTERS
        """
        self.packet = GPacketCountersTcp(conn, module_id, port_id, group_idx)
        """Packet counters"""
        self.payload = GPayloadCountersTcp(conn, module_id, port_id, group_idx)
        """Payload counters"""
        self.state = GStateCountersTcp(conn, module_id, port_id, group_idx)
        """State counters"""


class GRxHistogramTcp:
    """TCP RX Histogram"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.total_bytes = P4G_TCP_RX_TOTAL_BYTES_HIST(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_RX_TOTAL_BYTES_HIST
        """
        self.good_bytes = P4G_TCP_RX_GOOD_BYTES_HIST(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_RX_GOOD_BYTES_HIST
        """


class GTxHistogramTcp:
    """TCP TX Histogram"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.total_bytes = P4G_TCP_TX_TOTAL_BYTES_HIST(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_TX_TOTAL_BYTES_HIST
        """
        self.good_bytes = P4G_TCP_TX_GOOD_BYTES_HIST(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_TX_GOOD_BYTES_HIST
        """


class GConnHistogramTcp:
    """TCP Connection Histogram"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.establish_times = P4G_TCP_ESTABLISH_HIST(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_ESTABLISH_HIST
        """
        self.close_times = P4G_TCP_CLOSE_HIST(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_CLOSE_HIST
        """


class GHistogramTcp:
    """TCP histogram"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.connection = GConnHistogramTcp(conn, module_id, port_id, group_idx)
        """TCP Connection Histogram"""
        self.rx = GRxHistogramTcp(conn, module_id, port_id, group_idx)
        """TCP RX Histogram"""
        self.tx = GTxHistogramTcp(conn, module_id, port_id, group_idx)
        """TCP TX Histogram"""


class GReceiverWindowTcp:
    """TCP RWND Configuration"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.size = P4G_TCP_WINDOW_SIZE(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_WINDOW_SIZE
        """
        self.scaling = P4G_TCP_WINDOW_SCALING(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_WINDOW_SCALING
        """


class GCongestionWindowTcp:
    """TCP CWND Configuration"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.congestion_mode = P4G_TCP_CONGESTION_MODE(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_CONGESTION_MODE
        """
        self.icwnd_calc_method = P4G_TCP_ICWND_CALC_METHOD(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_ICWND_CALC_METHOD
        """


class GTcp:
    """TCP Stack Configuration"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.clear_post_test_statistics = P4G_CLEAR_POST_STAT(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_CLEAR_POST_STAT
        """
        self.rtt_value = P4G_TCP_RTT_VALUE(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_RTT_VALUE
        """
        self.iss_treshold = P4G_TCP_ISSTHRESH(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_TCP_ISSTHRESH
        """
        self.mss = GMaxSegmentSize(conn, module_id, port_id, group_idx)
        """Max segment size config
        
        :type: GMaxSegmentSize
        """
        self.rto = GRetransmissionTimeoutTcp(conn, module_id, port_id, group_idx)
        """Retransmission timeout config
        
        :type: GRetransmissionTimeoutTcp
        """
        self.ack = GAckTcp(conn, module_id, port_id, group_idx)
        """ACK config
        
        :type: GAckTcp
        """
        self.rwnd = GReceiverWindowTcp(conn, module_id, port_id, group_idx)
        """RWND config
        
        :type: GReceiverWindowTcp
        """
        self.cwnd = GCongestionWindowTcp(conn, module_id, port_id, group_idx)
        """CWND config
        
        :type: GCongestionWindowTcp
        """
        self.counters = GCountersTcp(conn, module_id, port_id, group_idx)
        """Counters.
        
        :type: GCountersTcp
        """
        self.histogram = GHistogramTcp(conn, module_id, port_id, group_idx)
        """Histogram.
        
        :type: GHistogramTcp
        """



from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands import (
    P4_ETH_TX_COUNTERS,
    P4_ETH_RX_COUNTERS,
    P4_PORT_TX_COUNTERS,
    P4_PORT_RX_COUNTERS,
    P4_PORT_COUNTERS,
    P4_IPV4_RX_COUNTERS,
    P4_IPV4_TX_COUNTERS,
    P4_IPV4_COUNTERS,
    P4_IPV6_RX_COUNTERS,
    P4_IPV6_TX_COUNTERS,
    P4_IPV6_COUNTERS,
    P4_ARP_RX_COUNTERS,
    P4_ARP_TX_COUNTERS,
    P4_ARP_COUNTERS,
    P4_NDP_RX_COUNTERS,
    P4_NDP_TX_COUNTERS,
    P4_NDP_COUNTERS,
    P4_ICMP_RX_COUNTERS,
    P4_ICMP_TX_COUNTERS,
    P4_ICMP_COUNTERS,
    P4_TCP_RX_COUNTERS,
    P4_TCP_TX_COUNTERS,
    P4_TCP_COUNTERS,
    P4_UDP_RX_COUNTERS,
    P4_UDP_TX_COUNTERS,
    P4_UDP_COUNTERS,
    P4_CLEAR_COUNTERS,
    P4_ETH_COUNTERS,
)


class PCountersIPv4:
    """IPV4 statistics"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = P4_IPV4_RX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_IPV4_RX_COUNTERS"""
        self.tx = P4_IPV4_TX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_IPV4_TX_COUNTERS"""
        self.total = P4_IPV4_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_IPV4_COUNTERS"""


class PCountersIPv6:
    """IPV6 statistics"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = P4_IPV6_RX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_IPV6_RX_COUNTERS"""
        self.tx = P4_IPV6_TX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_IPV6_TX_COUNTERS"""
        self.total = P4_IPV6_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_IPV6_COUNTERS"""


class PCountersICMP:
    """ICMP statistics"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = P4_ICMP_RX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_ICMP_RX_COUNTERS"""
        self.tx = P4_ICMP_TX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_ICMP_TX_COUNTERS"""
        self.total = P4_ICMP_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_ICMP_COUNTERS"""


class PCountersTCP:
    """ICMP statistics"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = P4_TCP_RX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_TCP_RX_COUNTERS"""
        self.tx = P4_TCP_TX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_TCP_TX_COUNTERS"""
        self.total = P4_TCP_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_TCP_COUNTERS"""


class PCountersUDP:
    """ICMP statistics"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = P4_UDP_RX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_UDP_RX_COUNTERS"""
        self.tx = P4_UDP_TX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_UDP_TX_COUNTERS"""
        self.total = P4_UDP_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_UDP_COUNTERS"""


class PCountersETH:
    """ICMP statistics"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = P4_ETH_RX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_ETH_RX_COUNTERS"""
        self.tx = P4_ETH_TX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_ETH_TX_COUNTERS"""
        self.total = P4_ETH_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_ETH_COUNTERS"""


class PCountersARP:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = P4_ARP_RX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_ARP_RX_COUNTERS"""
        self.tx = P4_ARP_TX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_ARP_TX_COUNTERS"""
        self.total = P4_ARP_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_ARP_COUNTERS"""


class PCountersNDP:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = P4_NDP_RX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_NDP_RX_COUNTERS"""
        self.tx = P4_NDP_TX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_NDP_TX_COUNTERS"""
        self.total = P4_NDP_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_NDP_COUNTERS"""


class PCounters:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.clear = P4_CLEAR_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_CLEAR_COUNTERS"""
        self.total_tx = P4_PORT_TX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_PORT_TX_COUNTERS"""
        self.total_rx = P4_PORT_RX_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_PORT_RX_COUNTERS"""
        self.total = P4_PORT_COUNTERS(conn, module_id, port_id)
        """
        :type: P4_PORT_COUNTERS"""

        self.ipv4 = PCountersIPv4(conn, module_id, port_id)
        """L47 port IPv4 counters."""
        self.ipv6 = PCountersIPv6(conn, module_id, port_id)
        """L47 port IPv6 counters."""
        self.icmp = PCountersICMP(conn, module_id, port_id)
        """L47 port ICMP counters."""
        self.tcp = PCountersTCP(conn, module_id, port_id)
        """L47 port TCP counters."""
        self.udp = PCountersUDP(conn, module_id, port_id)
        """L47 port UDP counters."""
        self.eth = PCountersETH(conn, module_id, port_id)
        """L47 port Ethernet counters."""
        self.arp = PCountersARP(conn, module_id, port_id)
        """L47 port ARP counters."""
        self.ndp = PCountersNDP(conn, module_id, port_id)
        """L47 port NDP counters."""



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands import (
    P4G_RAW_TEST_SCENARIO,
    P4G_RAW_PAYLOAD_TYPE,
    P4G_RAW_PAYLOAD_TOTAL_LEN,
    P4G_RAW_PAYLOAD,
    P4G_RAW_PAYLOAD_REPEAT_LEN,
    P4G_RAW_HAS_DOWNLOAD_REQ,
    P4G_RAW_CLOSE_CONN,
    P4G_RAW_UTILIZATION,
    P4G_RAW_DOWNLOAD_REQUEST,
    P4G_RAW_TX_DURING_RAMP,
    P4G_RAW_TX_TIME_OFFSET,
    P4G_RAW_BURSTY_TX,
    P4G_RAW_BURSTY_CONF,
    P4G_RAW_RX_PAYLOAD_LEN,
    P4G_RAW_REQUEST_REPEAT,
    P4G_RAW_CONN_INCARNATION,
    P4G_RAW_CONN_REPETITIONS,
    P4G_RAW_CONN_LIFETIME,
    P4G_APP_TRANSACTION_COUNTERS,
)


class GPayloadRaw:
    """Payload settings of Raw type of test application"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.type = P4G_RAW_PAYLOAD_TYPE(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_PAYLOAD_TYPE
        """

        self.total_length = P4G_RAW_PAYLOAD_TOTAL_LEN(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_PAYLOAD_TOTAL_LEN
        """

        self.rx_length = P4G_RAW_RX_PAYLOAD_LEN(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_RX_PAYLOAD_LEN
        """

        self.content = P4G_RAW_PAYLOAD(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_PAYLOAD
        """

        self.repeat_length = P4G_RAW_PAYLOAD_REPEAT_LEN(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_PAYLOAD_REPEAT_LEN
        """


class GConnectionRaw:
    """Connection settings of Raw type of test application"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.incarnation = P4G_RAW_CONN_INCARNATION(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_CONN_INCARNATION
        """

        self.repetitions = P4G_RAW_CONN_REPETITIONS(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_CONN_REPETITIONS
        """

        self.lifetime = P4G_RAW_CONN_LIFETIME(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_CONN_LIFETIME
        """

        self.close_condition = P4G_RAW_CLOSE_CONN(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_CLOSE_CONN
        """


class GBurstyRaw:
    """Burst settings of Raw type of test application"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.transmission = P4G_RAW_BURSTY_TX(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_BURSTY_TX
        """

        self.config = P4G_RAW_BURSTY_CONF(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_BURSTY_CONF
        """


class GTransmitRaw:
    """Transmission settings of Raw type of test application"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.during_ramp = P4G_RAW_TX_DURING_RAMP(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_TX_DURING_RAMP
        """

        self.time_offset = P4G_RAW_TX_TIME_OFFSET(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_TX_TIME_OFFSET
        """


class GDownloadRequestRaw:
    """Download request settings."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.server_must_wait = P4G_RAW_HAS_DOWNLOAD_REQ(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_HAS_DOWNLOAD_REQ
        """
        self.transactions_number = P4G_RAW_REQUEST_REPEAT(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_REQUEST_REPEAT
        """
        self.content = P4G_RAW_DOWNLOAD_REQUEST(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_DOWNLOAD_REQUEST
        """


class GCountersTransaction:
    """Transaction counters."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.transaction = P4G_APP_TRANSACTION_COUNTERS(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_APP_TRANSACTION_COUNTERS
        """


class GRaw:
    """Raw test application configuration."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.test_scenario = P4G_RAW_TEST_SCENARIO(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_TEST_SCENARIO
        """

        self.utilization = P4G_RAW_UTILIZATION(conn, module_id, port_id, group_idx)
        """
        
        :type: P4G_RAW_UTILIZATION
        """

        self.download_request = GDownloadRequestRaw(conn, module_id, port_id, group_idx)
        """Download request.
        
        :type: GDownloadRequestRaw
        """

        self.payload = GPayloadRaw(conn, module_id, port_id, group_idx)
        """Payload configuration.
        
        :type: GPayloadRaw
        """

        self.connection = GConnectionRaw(conn, module_id, port_id, group_idx)
        """Connection configuration.
        
        :type: GConnectionRaw
        """

        self.bursty = GBurstyRaw(conn, module_id, port_id, group_idx)
        """Burst configuration.
        
        :type: GBurstyRaw
        """

        self.tx = GTransmitRaw(conn, module_id, port_id, group_idx)
        """Transmit configuration.
        
        :type: GTransmitRaw
        """

        self.transaction_counter = GCountersTransaction(conn, module_id, port_id, group_idx)
        """Transaction counters.
        
        :type: GCountersTransaction
        """



from typing import (
    Iterable,
    Callable,
    Union,
    Type,
)
from loguru import logger
from .. import protocol

SYMBOL_REQUEST = "<G><w> -> </w></G>"
SYMBOL_PUSH = "<M><w> -P </w></M>"
SYMBOL_RESPONSE = "<Y><e> <- </e></Y>"


class TransportationLogger:
    __slots__ = ("debug", "identity_name", "__logger", "__state")

    def __init__(self, uid: str, debug: bool = False) -> None:
        self.debug = debug
        self.identity_name = f"bifrost-{uid}"
        self.__state: Union[Type[StateDebugOn], Type[StateDebugOff]] = StateDebugOff
        self.__logger = logger.bind(logger_name=self.identity_name)
        if self.debug:
            self.__logger.add(f"{self.identity_name}-info.log", rotation="100 MB", level="INFO")
            self.__logger.add(f"{self.identity_name}-debug.log", rotation="100 MB", level="DEBUG")
            self.__logger.add(f"{self.identity_name}-error.log", rotation="100 MB", level="ERROR")
            self.__logger.add(f"{self.identity_name}-warning.log", rotation="100 MB", level="WARNING")
            self.__state = StateDebugOn

    def _log(self, *args: Iterable[str], log_fn: Callable = logger.opt(colors=True).debug) -> None:
        for v in args:
            log_fn(v)

    def info(self, *args) -> None:
        self.__state.info(self, *args)

    def error(self, *args) -> None:
        self.__state.error(self, *args)

    def request_obj(self, request: protocol.Request) -> None:
        self.__state.request_obj(self, request)

    def response_obj(self, response: protocol.Response) -> None:
        self.__state.response_obj(self, response)

    def push_obj(self, response: protocol.Response) -> None:
        self.__state.push_obj(self, response)

    def draw_separator(self, symbol: str = "=") -> None:
        self.__state.draw_separator(self, symbol)


class StateDebugOn:
    @staticmethod
    def info(inst: "TransportationLogger", *args) -> None:
        inst._log(*args, log_fn=logger.info)

    @staticmethod
    def error(inst: "TransportationLogger", *args) -> None:
        inst._log(*args, log_fn=logger.error)

    @staticmethod
    def request_obj(inst: "TransportationLogger", request: protocol.Request) -> None:
        inst._log(f"{SYMBOL_REQUEST} {request!r}")

    @staticmethod
    def response_obj(inst: "TransportationLogger", response: protocol.Response) -> None:
        inst._log(f"{SYMBOL_RESPONSE} {response!r}")

    @staticmethod
    def push_obj(inst: "TransportationLogger", response: protocol.Response) -> None:
        inst._log(f"{SYMBOL_PUSH} {response!r}")

    @staticmethod
    def draw_separator(inst: "TransportationLogger", symbol: str) -> None:
        inst._log(symbol * 100, log_fn=logger.info)


class StateDebugOff:
    @staticmethod
    def info(inst: "TransportationLogger", *args) -> None:
        return None

    @staticmethod
    def error(inst: "TransportationLogger", *args) -> None:
        return None

    @staticmethod
    def request_obj(inst: "TransportationLogger", request: protocol.Request) -> None:
        return None

    @staticmethod
    def response_obj(inst: "TransportationLogger", response: protocol.Response) -> None:
        return None

    @staticmethod
    def push_obj(inst: "TransportationLogger", response: protocol.Response) -> None:
        return None

    @staticmethod
    def draw_separator(inst: "TransportationLogger", symbol: str) -> None:
        return None
